import zlib

from vstruct import VStruct,VArray
from vstruct.bitfield import *
from vstruct.primitives import *

'''
0 End
1 ShowFrame
2 DefineShape
4 PlaceObject
5 RemoveObject
6 DefineBits
7 DefineButton
8 JPEGTables
9 SetBackgroundColor
10 DefineFont
11 DefineText
12 DoAction
13 DefineFontInfo
14 DefineSound
15 StartSound
17 DefineButtonSound
18 SoundStreamHead
19 SoundStreamBlock
20 DefineBitsLossless
21 DefineBitsJPEG2
22 DefineShape2
23 DefineButtonCxform
24 Protect
26 PlaceObject2
28 RemoveObject2
32 DefineShape3
33 DefineText2
34 DefineButton2
35 DefineBitsJPEG3
36 DefineBitsLossless2
37 DefineEditText
39 DefineSprite
43 FrameLabel
45 SoundStreamHead2
46 DefineMorphShape
48 DefineFont2
56 ExportAssets
57 ImportAssets
58 EnableDebugger
59 DoInitAction
60 DefineVideoStream
61 VideoFrame
62 DefineFontInfo2
64 EnableDebugger2
65 ScriptLimits
66 SetTabIndex
69 FileAttributes
70 PlaceObject3
71 ImportAssets2
73 DefineFontAlignZones
74 CSMTextSettings
75 DefineFont3
76 SymbolClass
77 Metadata
78 DefineScalingGrid
82 DoABC
83 DefineShape4
84 DefineMorphShape2
86 DefineSceneAndFrameLabelData
87 DefineBinaryData
88 DefineFontName
89 StartSound2
90 DefineBitsJPEG4
91 DefineFont4
93 EnableTelemetry
'''

swftags = {
    0: 'End',
    1: 'ShowFrame',
    2: 'DefineShape',
    4: 'PlaceObject',
    5: 'RemoveObject',
    6: 'DefineBits',
    7: 'DefineButton',
    8: 'JPEGTables',
    9: 'SetBackgroundColor',
    10: 'DefineFont',
    11: 'DefineText',
    12: 'DoAction',
    13: 'DefineFontInfo',
    14: 'DefineSound',
    15: 'StartSound',
    17: 'DefineButtonSound',
    18: 'SoundStreamHead',
    19: 'SoundStreamBlock',
    20: 'DefineBitsLossless',
    21: 'DefineBitsJPEG2',
    22: 'DefineShape2',
    23: 'DefineButtonCxform',
    24: 'Protect',
    26: 'PlaceObject2',
    28: 'RemoveObject2',
    32: 'DefineShape3',
    33: 'DefineText2',
    34: 'DefineButton2',
    35: 'DefineBitsJPEG3',
    36: 'DefineBitsLossless2',
    37: 'DefineEditText',
    39: 'DefineSprite',
    41: 'SerialNumber',
    43: 'FrameLabel',
    45: 'SoundStreamHead2',
    46: 'DefineMorphShape',
    48: 'DefineFont2',
    56: 'ExportAssets',
    57: 'ImportAssets',
    58: 'EnableDebugger',
    59: 'DoInitAction',
    60: 'DefineVideoStream',
    61: 'VideoFrame',
    62: 'DefineFontInfo2',
    64: 'EnableDebugger2',
    65: 'ScriptLimits',
    66: 'SetTabIndex',
    69: 'FileAttributes',
    70: 'PlaceObject3',
    71: 'ImportAssets2',
    73: 'DefineFontAlignZones',
    74: 'CSMTextSettings',
    75: 'DefineFont3',
    76: 'SymbolClass',
    77: 'Metadata',
    78: 'DefineScalingGrid',
    82: 'DoABC',
    83: 'DefineShape4',
    84: 'DefineMorphShape2',
    86: 'DefineSceneAndFrameLabelData',
    87: 'DefineBinaryData',
    88: 'DefineFontName',
    89: 'StartSound2',
    90: 'DefineBitsJPEG4',
    91: 'DefineFont4',
    93: 'EnableTelemetry',
}


class RECT(VBitField):
    def __init__(self):
        VBitField.__init__(self)
        self.Nbits = v_bits(5)
        self.Xmin = v_bits(1)
        self.Xmax = v_bits(1)
        self.Ymin = v_bits(1)
        self.Ymax = v_bits(1)

    def pcb_Nbits(self):
        bitwidth = self.Nbits
        self['Xmin'].vsSetBitWidth(bitwidth)
        self['Xmax'].vsSetBitWidth(bitwidth)
        self['Ymin'].vsSetBitWidth(bitwidth)
        self['Ymax'].vsSetBitWidth(bitwidth)

class EncodedU32(v_number):

    def __init__(self):
        v_number.__init__(self)
        self._vs_fmt = None # to prevent misuse

    def vsParse(self, bytez, offset=0):
        self._vs_value = 0
        # this shit is *not* my fault... ;)
        for i in range(5):
            b = ord(bytez[offset + i])
            self._vs_value |= (b & 0x7f) << (7 * i)
            if not b & 0x80:
                break
        self._vs_length = i + 1
        return offset + self._vs_length

    def vsEmit(self):
        raise Exception('swf.EncodedU32() needs vsEmit!')

    def __repr__(self):
        return 'U32: %d' % self.vsGetValue()


class RGB(VStruct):
    def __init__(self):
        VStruct.__init__(self)
        self.Red    = v_uint8()
        self.Green  = v_uint8()
        self.Blue   = v_uint8()

class RGBA(VStruct):
    def __init__(self):
        VStruct.__init__(self)
        self.Red    = v_uint8()
        self.Green  = v_uint8()
        self.Blue   = v_uint8()
        self.Alpha  = v_uint8()

class MATRIX(VBitField):
    def __init__(self):
        VBitField.__init__(self)
        self.HasScale = v_bits(1)
        self.NScaleBits  = v_bits(5)
        self.ScaleX      = v_bits(1)
        self.ScaleY      = v_bits(1)
        self.HasRotate   = v_bits(1)
        self.NRotateBits = v_bits(5)
        self.RotateSkew0 = v_bits(1)
        self.RotateSkew1 = v_bits(1)
        self.NTransBits  = v_bits(5)
        self.TranslateX  = v_bits(1)
        self.TranslateY  = v_bits(1)

    def pcb_HasScale(self):
        if not self.HasScale:
            self.vsDelField('NScaleBits')
            self.vsDelField('ScaleX')
            self.vsDelField('ScaleY')

    def pcb_NScaleBits(self):
        bitwidth = self.NScaleBits
        self['ScaleX'].vsSetBitWidth(bitwidth)
        self['ScaleY'].vsSetBitWidth(bitwidth)

    def pcb_HasRotate(self):
        if not self.HasRotate:
            self.vsDelField('NRotateBits')
            self.vsDelField('RotateSkew0')
            self.vsDelField('RotateSkew1')

    def pcb_NRotateBits(self):
        bitwidth = self.NRotateBits
        self['RotateSkew0'].vsSetBitWidth(bitwidth)
        self['RotateSkew1'].vsSetBitWidth(bitwidth)

    def pcb_NTransBits(self):
        bitwidth = self.NTransBits
        self['TranslateX'].vsSetBitWidth(bitwidth)
        self['TranslateY'].vsSetBitWidth(bitwidth)

class GRADRECORD(VStruct):
    def __init__(self,shape=1):
        VStruct.__init__(self)
        self.Ratio  = v_uint8()
        if shape <= 2:
            self.Color = RGB()
        else:
            self.Color = RGBA()

class GRADIENT(VStruct):
    def __init__(self,shape=1):
        VStruct.__init__(self)
        self._swf_shape = shape
        self.BitFields       = VBitField()
        self.BitFields.SpreadMode           = v_bits(2)
        self.BitFields.InterpolationMode    = v_bits(2)
        self.BitFields.NumGradients         = v_bits(4)
        self.GradientRecords = VArray()

    def pcb_BitFields(self):
        gcount = self.BitFields.NumGradients
        for i in range(gcount):
            g = GRADRECORD(shape=self._swf_shape)
            self.GradientRecords.vsAddElement(g)

class FOCALGRADIENT(GRADIENT):
    def __init__(self,shape=4):
        GRADIENT.__init__(self)
        self.FocalPoint = FIXED8()

FILL_SOLID              = 0x00
FILL_LINEAR_GRADIENT    = 0x10
FILL_RADIAL_GRADIENT    = 0x12
FILL_FOCAL_GRADIENT     = 0x13
FILL_BITMAP_REPEAT      = 0x40
FILL_BITMAP_CLIPPED     = 0x41
FILL_BITMAP_NSMOOTH_REP = 0x42
FILL_BITMAP_NSMOOTH_CLP = 0x43

class FILLSTYLE(VStruct):
    def __init__(self, shape=1):
        VStruct.__init__(self)
        self._swf_shape = shape
        self.FillStyleType      = v_uint8()
        self.Color              = RGB()
        self.GradientMatrix     = MATRIX()
        self.Gradient           = GRADIENT(shape=shape)
        self.BitmapId           = v_uint16()
        self.BitmapMatrix       = MATRIX()

    def pcb_FillStyleType(self):
        fstype = self.FillStyleType
        if fstype != 0x00:
            self.vsDelField('Color')

        if fstype not in (0x10, 0x12, 0x13):
            self.vsDelField('GradientMatrix')
            self.vsDelField('Gradient')
        else:
            if fstype == 0x13:
                self.Gradient = FOCALGRADIENT()

        if fstype not in (0x40,0x41,0x42,0x43):
            self.vsDelField('BitmapId')
            self.vsDelField('BitmapMatrix')

class FILLSTYLEARRAY(VStruct):
    def __init__(self, shape=1):
        VStruct.__init__(self)
        self._swf_shape = shape
        self.FillStyleCount     = v_uint8()
        self.FillStyleCountEx   = v_uint16()
        self.FillStyles         = VArray()

    def pcb_FillStyleCount(self):
        if self.FillStyleCount != 0xff:
            self.vsDelField('FillStyleCountEx')
            self.swfAddFillStyles(self.FillStyleCount)

    def pcb_FillStyleCountEx(self):
        self.swfAddFillStyles(self.FillStyleCountEx)

    def swfAddFillStyles(self, count):
        for i in range(count):
            elem = FILLSTYLE(shape=self._swf_shape)
            self.FillStyles.vsAddElement(elem)

class LINESTYLE(VStruct):
    def __init__(self,shape=1):
        VStruct.__init__(self)
        self._swf_shape = shape
        self.Width  = v_uint16()
        colorcls = RGB
        if shape >= 3:
            colorcls = RGBA
        self.Color = colorcls()

class LINESTYLEARRAY(VStruct):
    def __init__(self,shape=1):
        VStruct.__init__(self)
        self._swf_shape = 1
        self.LineStyleCount     = v_uint8()
        self.LineStyleCountEx   = v_uint16()
        self.LineStyles         = VArray()

    def pcb_LineStyleCount(self):
        if self.LineStyleCount != 0xff:
            self.vsDelField('LineStyleCountEx')
            self.swfAddLineStyles(self.LineStyleCount)

    def pcb_LineStyleCountEx(self):
        self.swfAddLineStyles(self.LineStyleCountEx)

    def swfAddLineStyles(self, count):
        for i in range(count):
            elem = LINESTYLE(shape=self._swf_shape)
            self.LineStyles.vsAddElement(elem)

class SHAPEWITHSTYLE(VStruct):
    def __init__(self, shape=1):
        VStruct.__init__(self)
        self.FillStyles     = FILLSTYLEARRAY(shape=shape)
        self.LineStyles     = LINESTYLEARRAY(shape=shape)
        self.NumBits        = VBitField()
        self.NumBits.NumFillBits    = v_bits(4)
        self.NumBits.NumLineBits    = v_bits(4)
        self.ShapeRecords           = VArray()

class SwfColorTransform(VBitField):
    def __init__(self):
        VBitField.__init__(self)
        self.HasAddTerms    = v_bits(1)
        self.HasMultTerms   = v_bits(1)
        self.Nbits          = v_bits(4)
        self.RedMultTerm    = v_bits(1)
        self.GreenMultTerm  = v_bits(1)
        self.BlueMultTerm   = v_bits(1)
        self.RedAddTerm     = v_bits(1)
        self.GreenAddTerm   = v_bits(1)
        self.BlueAddTerm    = v_bits(1)

    def pcb_Nbits(self):
        bitwidth = self.Nbits

        if not self.HasAddTerms:
            self.vsDelField('RedAddTerm')
            self.vsDelField('GreenAddTerm')
            self.vsDelField('BlueAddTerm')
        else:
            self['RedAddTerm'].vsSetBitWidth(bitwidth)
            self['GreenAddTerm'].vsSetBitWidth(bitwidth)
            self['BlueAddTerm'].vsSetBitWidth(bitwidth)

        if not self.HasMultTerms:
            self.vsDelField('RedMultTerm')
            self.vsDelField('GreenMultTerm')
            self.vsDelField('BlueMultTerm')
        else:
            self['RedMultTerm'].vsSetBitWidth(bitwidth)
            self['GreenMultTerm'].vsSetBitWidth(bitwidth)
            self['BlueMultTerm'].vsSetBitWidth(bitwidth)

class SwfTagCodeLength(VBitField):
    def __init__(self):
        VBitField.__init__(self)
        self.TagCode = v_bits(10)
        self.TagLength = v_bits(6)

class SwfTagHeader(VStruct):
    def __init__(self):
        VStruct.__init__(self)
        self.TagCodeAndLength = v_uint16()
        self.OptionalLength   = v_uint32()

    def __repr__(self):
        tagcode = self.getSwfTagCode()
        taglen = self.getSwfTagLength()
        tagname = swftags.get(tagcode,'Unknown')
        return '%s (%d)' % (tagname,taglen)

    def pcb_TagCodeAndLength(self):
        if self.TagCodeAndLength & 0x3f != 0x3f:
            self.vsDelField('OptionalLength')

    def getSwfTagCode(self):
        return self.TagCodeAndLength >> 6

    def getSwfTagLength(self):
        taglen = self.TagCodeAndLength & 0x3f
        if taglen == 0x3f:
            taglen = self.OptionalLength
        return taglen

class PlaceObject(VStruct):
    def __init__(self, size):
        VStruct.__init__(self)
        self.CharacterId    = v_uint16()
        self.Depth          = v_uint16()
        self.Matrix         = MATRIX()
        self.ColorTransform = SwfColorTransform()

        self._swf_tagsize = size

    def vsParse(self, bytez, offset=0):
        # we *only* do this this way because we dont know
        # how big the Matrix field is until we parse it...
        offset = self['CharacterId'].vsParse(bytez,offset=offset)
        offset = self['Depth'].vsParse(bytez,offset=offset)
        offset = self['Matrix'].vsParse(bytez,offset=offset)
        if offset >= self._swf_tagsize:
            self.vsDelField('ColorTransform')
        else:
            offset = self['ColorTransform'].vsParse(bytez,offset=offset)
        return offset

class FileAttributes(VBitField):
    def __init__(self, size):
        VBitField.__init__(self)
        self.Reserved0      = v_bits(1)
        self.UseDirectBlit  = v_bits(1)
        self.UseGPU         = v_bits(1)
        self.HasMetadata    = v_bits(1)
        self.ActionScript3  = v_bits(1)
        self.Reserved1      = v_bits(2)
        self.UseNetwork     = v_bits(1)
        self.ReservedEnd    = v_bits(24)

class SetBackgroundColor(VStruct):
    def __init__(self, size):
        VStruct.__init__(self)
        self.BackgroundColor = RGB()

class OffsetName(VStruct):
    def __init__(self):
        VStruct.__init__(self)
        self.Offset     = EncodedU32()
        self.Name       = v_zstr()

class DefineSceneAndFrameLabelData(VStruct):
    def __init__(self, size):
        VStruct.__init__(self)
        self.SceneCount         = EncodedU32()
        self.Scenes             = VArray()
        self.FrameLabelCount    = EncodedU32()
        self.Frames             = VArray()

    def pcb_SceneCount(self):
        count = self.SceneCount
        self.Scenes.vsAddElements(count, OffsetName)

    def pcb_FrameLabelCount(self):
        count = self.FrameLabelCount
        self.Scenes.vsAddElements(count, OffsetName)

class DefineShape4(VStruct):
    def __init__(self, size):
        VStruct.__init__(self)
        self.ShapeId        = v_uint16()
        self.ShapeBounds    = RECT()
        self.EdgeBounds     = RECT()
        self.ShapeFlags     = VBitField()
        self.Shapes         = SHAPEWITHSTYLE(shape=4)

        self.ShapeFlags.Reserved                = v_bits(5)
        self.ShapeFlags.UsesFillWindingRule     = v_bits(1)
        self.ShapeFlags.UsesNonScalingStrokes   = v_bits(1)
        self.ShapeFlags.UsesScalingStrokes      = v_bits(1)

class SwfTag(VStruct):
    def __init__(self):
        VStruct.__init__(self)
        self.TagHeader = SwfTagHeader()
        self.TagData = VStruct()

    def pcb_TagHeader(self):
        taglen = self.TagHeader.getSwfTagLength()
        tagcode = self.TagHeader.getSwfTagCode()
        tagname = swftags.get(tagcode,'UnknownTag')
        tagclass = globals().get(tagname,v_bytes)
        self.TagData = tagclass(taglen)

class SwfHeader(VStruct):
    def __init__(self):
        VStruct.__init__(self)
        self.Signature  = v_str(3)
        self.Version    = v_uint8()
        self.FileLength = v_uint32()
        self.FrameSize  = RECT()
        self.FrameRate  = v_uint16()
        self.FrameCount = v_uint16()

class SwfFile(VStruct):

    def __init__(self):
        VStruct.__init__(self)
        self.Header = SwfHeader()
        self.Tags = VArray()

    def vsParse(self, bytez, offset=0):
        magic = bytez[offset:offset+3]
        if magic not in ('FWS','CWS','ZWS'):
            raise Exception('Invalid Swf Magic')

        if magic == 'CWS':
            bytez = bytez[:8] + zlib.decompress(bytez[8:])

        offset = self.Header.vsParse(bytez, offset=offset)
        i = 0
        while offset < len(bytez):
            swftag = SwfTag()
            offset = swftag.vsParse(bytez, offset=offset)
            self.Tags.vsAddElement(swftag)
            i += 1
            if i > 3:
                break



import logging

import envi.threads as e_thread

# Some common GUI helpers
from PyQt5 import QtCore 
from PyQt5.QtGui import QRegExpValidator
from PyQt5.QtWidgets import QTreeView, QDialog, QLineEdit, QComboBox, QVBoxLayout, QHBoxLayout, QDialogButtonBox, QLabel, QMessageBox
from vqt.main import idlethread

logger = logging.getLogger(__name__)

if not len(logger.handlers):
    logger.addHandler(logging.StreamHandler())


class ACT:
    def __init__(self, meth, *args, **kwargs):
        self.meth = meth
        self.args = args
        self.kwargs = kwargs
        self.newthread = False

    def setNewThread(self, newthread=True):
        self.newthread = newthread

    def __call__(self):
        if self.newthread:
            meth = e_thread.firethread(self.meth)
        else:
            meth = self.meth

        try:
            return meth( *self.args, **self.kwargs )

        except Exception as e:
            logger.warning("error in ACT(%s, %s, %s)", str(self.meth), str(self.args), str(self.kwargs))
            logger.warning(str(e))
            logger.info("", exc_info=1)


class VqtModel(QtCore.QAbstractItemModel):

    columns = ('one','two')
    editable = None
    dragable = False

    def __init__(self, rows=()):
        QtCore.QAbstractItemModel.__init__(self)
        # Make sure the rows are lists ( so we can mod them )
        self.rows = [ list(row) for row in rows ]
        if self.editable is None:
            self.editable = [False,] * len(self.columns)

    def index(self, row, column, parent):
        return self.createIndex(row, column, self.rows[row])

    def parent(self, index):
        return QtCore.QModelIndex()

    def rowCount(self, index):
        if index.internalPointer() in self.rows:
            return 0
        return len(self.rows)

    def data(self, index, role):
        if role == 0:
            row = index.row()
            col = index.column()
            return self.rows[row][col]

        else:
            return None

    def columnCount(self, index):
        return len(self.columns)

    def headerData(self, column, orientation, role):
        if (orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole):
            return self.columns[column]
        return None

    def flags(self, index):
        if not index.isValid():
            return 0
        flags = QtCore.QAbstractItemModel.flags(self, index)
        col = index.column()
        if self.editable[col]:
            flags |= QtCore.Qt.ItemIsEditable

        if self.dragable:
            flags |= QtCore.Qt.ItemIsDragEnabled# | QtCore.Qt.ItemIsDropEnabled

        return flags

    #def data(self, index, role):
        #if not index.isValid():
            #return None
        #item = index.internalPointer()
        #if role == QtCore.Qt.DisplayRole:
            #return item.data(index.column())
        #if role == QtCore.Qt.UserRole:
            #return item
        #return None

    #def _vqt_set_data(self, row, col, value):
        #return False

    #def appends(self, rows):

    def append(self, row, parent=QtCore.QModelIndex()):
        #pidx = self.createIndex(parent.row(), 0, parent)
        i = len(self.rows)
        self.beginInsertRows(parent, i, i)
        self.rows.append( row )
        #node = parent.append(rowdata)
        self.endInsertRows()
        self.layoutChanged.emit()

    def setData(self, index, value, role=QtCore.Qt.EditRole):

        if not index.isValid():
            return False

        # If this is the edit role, fire the vqEdited thing
        if role == QtCore.Qt.EditRole:
            #value = self.vqEdited(node, index.column(), value)
            #if value is None:
                #return False

            row = index.row()
            col = index.column()
            if not self._vqt_set_data( row, col, value ):
                return False

        return True

    def pop(self, row, parent=QtCore.QModelIndex()):
        self.beginRemoveRows(parent, row, row+1)
        self.rows.pop(row)
        self.endRemoveRows()

    #def mimeTypes(self):
        #types = QtCore.QStringList()
        #types.append('vqt/row')
        #return types

    #def mimeData(self, idx):
        #nodes = [ self.rows[i.row()][-1] for i in idx ]
        #mdata = QtCore.QMimeData()
        #mdata.setData('vqt/rows',json.dumps(nodes))
        #return mdata

class VqtView(QTreeView):

    def __init__(self, parent=None):
        QTreeView.__init__(self, parent=parent)
        self.setAlternatingRowColors( True )
        self.setSortingEnabled( True )

    def getSelectedRows(self):
        ret = []
        rdone = {}
        model = self.model()
        for idx in self.selectedIndexes():

            if rdone.get(idx.row()):
                continue

            rdone[idx.row()] = True
            ret.append(model.mapToSource(idx).internalPointer())

        return ret

    def setModel(self, model):
        smodel = QtCore.QSortFilterProxyModel(parent=self)
        smodel.setSourceModel(model)
        ret = QTreeView.setModel(self, smodel)
        c = len(model.columns)
        for i in range(c):
            self.resizeColumnToContents(i)
        return ret

    def getModelRows(self):
        return self.model().sourceModel().rows

    def getModelRow(self, idx):
        idx = self.model().mapToSource(idx)
        return idx.row(), idx.internalPointer()


class DynamicDialog(QDialog):
    '''
    Pop up and ask the questions.  If "OK", returns a dict with answers.  
    Intended for easy access by extensions, and possibly for envi.Config items.
    eg:

    >>> dynd = vcmn.DynamicDialog('Test Dialog')
    >>> dynd.addComboBox('testbox', ["a", 'b', 'c'], dfltidx=2)
    >>> dynd.addTextField('foo', dflt="blah blah")
    >>> dynd.addIntHexField('bar', dflt=47145)
    >>> results = dynd.prompt()
        <when returns, after user clicks OK>
    >>> print(results)
    {'testbox': 'c', 'foo': 'blah blah', 'bar': 47145}

            or...

    >>> results = dynd.prompt()
        <if user hits cancel>
    >>> print(results)
    {}

    '''
    _TEXT = 0
    _COMBO = 1
    _INTHEX = 2

    def __init__(self, title, height=100, width=300, parent=None):
        super().__init__(parent)

        self.setWindowTitle(title)
        self.items = {}
        self.resize(width, height)
        self.vbox = QVBoxLayout()
        self.setLayout(self.vbox)
        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel);
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

    def addComboBox(self, fieldname, itemlist, dfltidx=None, title=None):
        '''
        Adds a Combo Box to the dialog.  The returned value will be one of the 
        options provided in itemlist.

        dfltidx is the index in the itemlist which should start out selected.
        default is the first item.
        '''
        if fieldname in self.items:
            raise Exception("ComboBox: Adding field to DynamicDialog twice: %r (existing: %r)"\
                    % (fieldname, self.items.get(fieldname)))

        cb = QComboBox()
        cb.addItems(itemlist)
        self.items[fieldname] = (self._COMBO, cb)

        if dfltidx is not None:
            cb.setCurrentIndex(dfltidx)
        if title is None:
            title = fieldname

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel(title))
        hbox.addWidget(cb)
        self.vbox.addLayout(hbox)

    def addTextField(self, fieldname, dflt=None, title=None):
        '''
        Adds a standard Text Box (QLineEdit) to the dialog.
        Return value is basically untouched.
        '''
        if fieldname in self.items:
            raise Exception("Text: Adding field to DynamicDialog twice: %r (existing: %r)"\
                    % (fieldname, self.items.get(fieldname)))

        le = QLineEdit()
        self.items[fieldname] = (self._TEXT, le)

        if dflt is not None:
            le.setText(dflt)
        if title is None:
            title = fieldname

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel(title))
        hbox.addWidget(le)
        self.vbox.addLayout(hbox)

    def addIntHexField(self, fieldname, dflt=None, title=None):
        '''
        Adds a number field to the dialog.  It's a QLineEdit field with a 
        QRegExpValidator, using regex for Hexidecimal numbers, which also
        allows for Decimal.  
        Returned value is an int.
        The value is interpreted as decimal (ie. int(foo)) unless non-numeric digits 
        encountered, in which case it's interpreted as hex (ie. int(foo, 16))
        '''
        if fieldname in self.items:
            raise Exception("IntHex: Adding field to DynamicDialog twice: %r (existing: %r)"\
                    % (fieldname, self.items.get(fieldname)))

        le = QLineEdit()
        le.setValidator(QRegExpValidator(QtCore.QRegExp("^(-)?(0x)?[0-9a-fA-F]+$")))
        self.items[fieldname] = (self._INTHEX, le)

        if dflt is not None:
            le.setText(str(dflt))
        if title is None:
            title = fieldname

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel(title))
        hbox.addWidget(le)
        self.vbox.addLayout(hbox)

    def prompt(self):
        '''
        Temporarily adds a ButtonBox with "OK" and "Cancel", and shows the 
        dialog to the user.  Then removes the buttonBox.
        '''
        self.vbox.addWidget(self.buttonBox);
        res = self.exec_()
        self.vbox.removeWidget(self.buttonBox);

        retval = {}
        if res:
            for key, (ftype, field) in self.items.items():
                if ftype == self._INTHEX:
                    val = field.text()
                    if len(val):
                        try:
                            val = int(val)
                        except ValueError:
                            val = int(val, 16)
                    else:
                        val = 0

                elif ftype == self._TEXT:
                    val = field.text()

                elif ftype == self._COMBO:
                    val = field.currentText()

                retval[key] = val

        return retval

@idlethread
def warning(msg, info):
    msgbox = QMessageBox()
    msgbox.setWindowTitle('Warn: %s' % msg)
    msgbox.setText('Warn: %s' % msg)
    msgbox.setInformativeText(info)
    msgbox.setIcon(QMessageBox.Warning)
    msgbox.exec_()

@idlethread
def information(msg, info):
    msgbox = QMessageBox()
    msgbox.setWindowTitle('%s' % msg)
    msgbox.setText('%s' % msg)
    msgbox.setInformativeText(info)
    msgbox.setIcon(QMessageBox.Information)
    msgbox.exec_()

@idlethread
def scripterr(msg, info):
    msgbox = QMessageBox()
    msgbox.setWindowTitle('Script Error: %s' % msg)
    msgbox.setText('Script Error: %s' % msg)
    msgbox.setInformativeText(info)
    msgbox.exec_()




import vstruct

from vstruct.primitives import *


#const tag descriptions from wikipedia
#Tag byte    Additional bytes    Description of constant
#1   2+x bytes utf-8 string
#3   4 bytes Integer: a signed 32-bit two's complement number in big-endian format
#4   4 bytes Float: a 32-bit single-precision IEEE 754 floating-point number
#5   8 bytes Long: a signed 64-bit two's complement number in big-endian format (takes two slots in the constant pool table)
#6   8 bytes Double: a 64-bit double-precision IEEE 754 floating-point number (takes two slots in the constant pool table)
#7   2 bytes Class reference: an index within the constant pool to a UTF-8 string containing the fully qualified class name (in internal format)
#8   2 bytes String reference: an index within the constant pool to a UTF-8 string
#9   4 bytes Field reference: two indexes within the constant pool, the first pointing to a Class reference, the second to a Name and Type descriptor.
#10  4 bytes Method reference: two indexes within the constant pool, the first pointing to a Class reference, the second to a Name and Type descriptor.
#11  4 bytes Interface method reference: two indexes within the constant pool, the first pointing to a Class reference, the second to a Name and Type descriptor.
#12  4 bytes Name and type descriptor: two indexes to UTF-8 strings within the constant pool, the first representing a name (identifier) and the second a specially encoded type descriptor.

class ConstPoolStr(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.size       = v_uint16(bigend=True)
        self.strbytes   = v_wstr(encode='utf8')

    def pcb_size(self):
        self.vsGetField('strbytes').vsSetLength( self.size )

tag_classes = {
    1:  ConstPoolStr,
    3:  v_uint32,
    4:  v_uint32,
    5:  v_uint64,
    6:  v_uint64,
    7:  v_uint16,
    8:  v_uint16,
    9:  v_uint32,
    10: v_uint32,
    11: v_uint32,
    12: v_uint32,
}

# Constant Tag Types
CONSTANT_Utf8               = 1
CONSTANT_Integer            = 3
CONSTANT_Float              = 4
CONSTANT_Long               = 5
CONSTANT_Double             = 6
CONSTANT_Class              = 7
CONSTANT_String             = 8
CONSTANT_Fieldref           = 9
CONSTANT_Methodref          = 10
CONSTANT_InterfaceMethodref = 11
CONSTANT_NameAndType        = 12
CONSTANT_MethodHandle       = 15
CONSTANT_MethodType         = 16
CONSTANT_InvokeDynamic      = 18

# Access Flags Values
ACC_PUBLIC      = 0x0001  # Declared public; may be accessed from outside its package.
ACC_PRIVATE     = 0x0002  # Declared private; usable only within the defining class.
ACC_PROTECTED   = 0x0004  # Declared protected; may be accessed within subclasses.
ACC_STATIC      = 0x0008  # Declared static.
ACC_FINAL       = 0x0010  # Declared final; never directly assigned to after object construction.
ACC_VOLATILE    = 0x0040  # Declared volatile; cannot be cached.
ACC_TRANSIENT   = 0x0080  # Declared transient; not written or read by a persistent object manager.
ACC_SYNTHETIC   = 0x1000  # Declared synthetic; not present in the source code.
ACC_ENUM        = 0x4000  # Declared as an element of an enum.

class ConstPoolInfo(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.tag    = v_uint8()
        self.data   = vstruct.VStruct()

    def pcb_tag(self):
        cls = tag_classes.get(self.tag)
        if cls is None:
            raise Exception('Unknown ConstPoolInfo Tag: %s' % self.tag )
        self.data.tagval = cls()

class AttributeInfo(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.attribute_name_index   = v_uint16(bigend=True)
        self.attribute_length       = v_uint32(bigend=True)
        self.attribute              = v_bytes()

    def pcb_attribute_length(self):
        self.vsGetField('attribute').vsSetLength( self.attribute_length )

class FieldInfo(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.access_flags       = v_uint16(bigend=True)
        self.name_index         = v_uint16(bigend=True)
        self.descriptor_index   = v_uint16(bigend=True)
        self.attributes_count   = v_uint16(bigend=True)
        self.attributes         = vstruct.VArray()

    def pcb_attributes_count(self):
        self.attributes.vsAddElements( self.attributes_count, AttributeInfo )

class MethodInfo(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.access_flags       = v_uint16(bigend=True)
        self.name_index         = v_uint16(bigend=True)
        self.descriptor_index   = v_uint16(bigend=True)
        self.attributes_count   = v_uint16(bigend=True)
        self.attributes         = vstruct.VArray()

    def pcb_attributes_count(self):
        self.attributes.vsAddElements( self.attributes_count, AttributeInfo )

class JavaClass(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.magic          = v_uint32(bigend=True)
        self.minor_versino  = v_uint16(bigend=True)
        self.major_version  = v_uint16(bigend=True)

        self.const_pool_cnt = v_uint16(bigend=True)
        self.const_pool     = vstruct.VArray()

        self.access_flags   = v_uint16(bigend=True)
        self.this_class     = v_uint16(bigend=True)
        self.super_class    = v_uint16(bigend=True)

        self.interface_cnt  = v_uint16(bigend=True)
        self.interfaces     = vstruct.VArray()

        self.fields_cnt     = v_uint16(bigend=True)
        self.fields         = vstruct.VArray()

        self.methods_cnt    = v_uint16(bigend=True)
        self.methods        = vstruct.VArray()

        self.attributes_cnt = v_uint16(bigend=True)
        self.attributes     = vstruct.VArray()

    def pcb_const_pool_cnt(self):
        # Count is off by one according to the spec
        self.const_pool.vsAddElements( self.const_pool_cnt - 1, ConstPoolInfo )

    def pcb_interface_cnt(self):
        for i in range( self.interface_cnt ):
            self.interfaces.vsAddElement( v_uint16( bigend=True ) )

    def pcb_fields_cnt(self):
        self.fields.vsAddElements( self.fields_cnt, FieldInfo )

    def pcb_methods_cnt(self):
        self.methods.vsAddElements( self.methods_cnt, MethodInfo )

    def pcb_attributes_cnt(self):
        self.attributes.vsAddElements( self.attributes_cnt, AttributeInfo )

    def getClassName(self):
        return self.const_pool[ self.this_class ].data.tagval.strbytes

    def getSuperClassName(self):
        return self.const_pool[ self.super_class ].data.tagval.strbytes

    def getClassFields(self):
        '''
        Get the fields defined by this class as a tuple of
        ( fieldname, fieldtype, attribs ) where attribs is a dict.
        '''
        ret = []
        for fname,fieldinfo in self.fields:

            fieldname = self.const_pool[ fieldinfo.name_index - 1 ].data.tagval.strbytes
            descname = self.const_pool[ fieldinfo.descriptor_index - 1 ].data.tagval.strbytes

            attrs = {}
            for afield, attrinfo in fieldinfo.attributes:
                attrname = self.const_pool[ attrinfo.attribute_name_index - 1 ].data.tagval.strbytes
                attrs[ attrname ] = attrinfo.attribute

            ret.append( (fieldname, descname, attrs) )
        return ret

    def getClassMethods(self):
        ret = []
        for fname,methinfo in self.methods:
            methname = self.const_pool[ methinfo.name_index - 1 ].data.tagval.strbytes
            attrs = {}
            for afield, attrinfo in methinfo.attributes:
                attrname = self.const_pool[ attrinfo.attribute_name_index - 1 ].data.tagval.strbytes
                attrs[ attrname ] = attrinfo.attribute
            ret.append( (methname, attrs) )
        return ret

    def getClassAttributes(self):
        attrs = {}
        for afield, attrinfo in self.attributes:
            attrname = self.const_pool[ attrinfo.attribute_name_index - 1 ].data.tagval.strbytes
            attrs[ attrname ] = attrinfo.attribute
        return attrs



"""
Test for platform functionality (for internal use).
"""
# Copyright (C) 2007 Invisigoth - See LICENSE file for details
import vtrace
import vtrace.platforms.base as v_base

############################################
#
# FIXME this is dorked for now based on the new platforms/archs design
#
############################################


def auditTracer(trace):
    """
    Print out a list of platform requirements and whether
    a particular tracer meets them.  This is mostly a
    development tool to determine what's left to do on a
    tracer implementation.
    """
    for mname in dir(v_base.BasePlatformMixin):
        if "__" in mname:
            continue
        if getattr(trace.__class__, mname) == getattr(v_base.BasePlatformMixin, mname):
            print("LACKS: %s" % mname)
        else:
            print("HAS: %s" % mname)


if __name__ == "__main__":
    trace = vtrace.getTrace()
    auditTracer(trace)




"""
An i386 specific function analysis module that is designed to
attempt to detect the calling convention.
"""
import collections

import vivisect.impemu.monitor as viv_imp_monitor

import vivisect.exc as v_exc
from vivisect.const import *

import vivisect.analysis.generic.switchcase as vag_switch

import envi.archs.i386 as e_i386
import envi.archs.i386.opconst as e_i386const

regcalls = {
    (e_i386.REG_ECX,):               ('thiscall', 1),
    (e_i386.REG_EAX,):               ('bfastcall', 1),
    (e_i386.REG_EAX, e_i386.REG_EDX): ('bfastcall', 2),
    (e_i386.REG_EAX, e_i386.REG_ECX): ('bfastcall', 3),
    (e_i386.REG_ECX, e_i386.REG_EDX): ('msfastcall', 2),
    (e_i386.REG_EAX, e_i386.REG_ECX, e_i386.REG_EDX): ('bfastcall', 3),
}

# Arrange the same data for a name lookup

empty = collections.defaultdict(lambda x: ('int','arg%d' % x))
argnames = {
    'thiscall':     {0: ('void *','ecx'),},
    'msfastcall':   {0: ('int','ecx'), 1: ('int','edx')},
    'bfastcall':    {0: ('int','eax'), 1: ('int','edx'), 2: ('int','ecx'),},
}

def argcname(callconv, idx):
    ret = argnames.get(callconv,empty).get(idx)
    if ret is None:
        ret = ('int','arg%d' % idx)
    return ret

class AnalysisMonitor(viv_imp_monitor.AnalysisMonitor):

    def __init__(self, vw, fva):
        viv_imp_monitor.AnalysisMonitor.__init__(self, vw, fva)
        self.retbytes = None
        self.endstack = None
        self.addDynamicBranchHandler(vag_switch.analyzeJmp)
        self.badops = vw.arch.archGetBadOps()

    def prehook(self, emu, op, starteip):
        if op in self.badops:
            raise v_exc.BadOpBytes(op.va)

        viv_imp_monitor.AnalysisMonitor.prehook(self, emu, op, starteip)

        if op.opcode == e_i386const.INS_LEA:    # x86 only
            i = 1
            o = op.opers[i]
            discrete = o.isDiscrete()
            operva = o.getOperAddr(op, emu)
            # keep track of the max here, but save it for later too...
            stackoff = emu.getStackOffset(operva)
            if stackoff and stackoff >= 0:
                self.stackmax = max(self.stackmax, stackoff)
                self.stackargs[stackoff] = True

            self.operrefs.append((starteip, i, operva, o.tsize, stackoff, discrete))


        # Do return related stuff before we execute the opcode
        if op.isReturn():
            self.endstack = emu.getStackCounter()
            if len(op.opers):
                self.retbytes = op.opers[0].imm


def buildFunctionApi(vw, fva, emu, emumon, stkstart):
    # More than 40 args?  no way...
    argc = stackargs = (int(emumon.stackmax) >> 2)
    if argc > 40:
        emumon.logAnomaly(emu, fva, 'Crazy Stack Offset Touched: 0x%.8x' % emumon.stackmax)
        argc = 0

    callconv = "cdecl"  # Default to cdecl
    # see if we have stdcall return bytes
    if emumon.retbytes is not None:
        callconv = "stdcall"
        argc = emumon.retbytes >> 2

    stackidx = 0  # arg index of first *stack* arg

    # Log registers we used but didn't init
    # but don't take into account ebp and esp
    emu.uninit_use.pop(e_i386.REG_ESP, None)
    emu.uninit_use.pop(e_i386.REG_EBP, None)
    undefkeys = list(emu.uninit_use.keys())
    undefkeys.sort()

    undeflen = len(undefkeys)
    if undeflen:
        regcall = regcalls.get(tuple(undefkeys))
        if regcall is not None:
            callconv, addargc = regcall
            argc += addargc

        vw.setFunctionMeta(fva, "UndefRegUse", undefkeys)

    # if we're callee cleanup, make sure we *actually* clean up our space
    # otherwise, revert us to caller cleanup
    realcallconv = callconv
    if emumon.endstack:
        stkoff = (emumon.endstack - stkstart) >> 2
        if callconv in argnames:
            # do our stack args line up with what we cleaned up?
            if abs(stkoff) != stackargs:
                # we're probably caller cleanup then
                realcallconv = callconv + '_caller'

    if argc > 64:
        callconv = 'unkcall'
        argc = 0
    # Add argument indexes to our argument names
    funcargs = [ argcname(callconv, i) for i in range(argc) ]
    api = ('int',None,realcallconv,None,funcargs)

    vw.setFunctionApi(fva, api)
    return api

def analyzeFunction(vw, fva):

    emu = vw.getEmulator(va=fva)
    emumon = AnalysisMonitor(vw, fva)

    stkstart = emu.getStackCounter()
    emu.setEmulationMonitor(emumon)
    emu.runFunction(fva, maxhit=1)

    # Do we already have API info in meta?
    # NOTE: do *not* use getFunctionApi here, it will make one!
    api = vw.getFunctionMeta(fva, 'api')
    if api is None:
        api = buildFunctionApi(vw, fva, emu, emumon, stkstart)

    rettype,retname,callconv,callname,callargs = api
    if callconv == 'unkcall':
        return

    argc = len(callargs)
    cc = emu.getCallingConvention(callconv)
    stcount = cc.getNumStackArgs(emu, argc)
    stackidx = argc - stcount
    baseoff = cc.getStackArgOffset(emu, argc)

    # Register our stack args as function locals
    for i in range(stcount):
        vw.setFunctionLocal(fva, baseoff + ( i * 4 ), LSYM_FARG, i+stackidx)

    emumon.addAnalysisResults(vw, emu)



from PyQt5 import QtCore, QtGui
from PyQt5.QtWidgets import *

import vtrace
from vtrace.const import *

import envi.qt as envi_qt
import envi.qt.memory as envi_qt_memory
import envi.qt.memorymap as envi_qt_memmap
import envi.memcanvas as e_mcanv
import envi.memcanvas.renderers as e_mem_rend

import vqt.tree as vq_tree
import vqt.colors as vq_colors

from vqt.main import workthread, idlethread, idlethreadsync

'''
QtGui objects which assist in GUIs which use vtrace parts.
'''

class VQTraceNotifier(vtrace.Notifier):
    '''
    A bit of shared mixin code for the handling of vtrace
    notifier callbacks in various VQTreeViews...
    '''
    def __init__(self, trace=None):
        self.trace = trace
        vtrace.Notifier.__init__(self)
        if trace is not None:
            self.trace.registerNotifier(NOTIFY_ALL, self)

    @idlethreadsync
    # FIXME this should be part of a shared API!
    def notify(self, event, trace):
        if event in [NOTIFY_CONTINUE, NOTIFY_DETACH, NOTIFY_EXIT]:
            self.setEnabled(False)

        else:
            # If the trace is just going to run again, skip the update.
            if not trace.shouldRunAgain():
                self.setEnabled(True)
                self.vqLoad()

class RegisterListModel(envi_qt_memory.EnviNavModel):
    columns = ('Name', 'Hex', 'Dec', 'Best')
    editable = [False, True, True, False]
    register_edited = QtCore.pyqtSignal(QtCore.QModelIndex, QtCore.QVariant)

    def __init__(self, parent=None):
        envi_qt_memory.EnviNavModel.__init__(self, 0, parent=parent)

    def data(self, index, role):
        if not index.isValid():
            return None

        item = index.internalPointer()
        if role == QtCore.Qt.DisplayRole or role == QtCore.Qt.EditRole:
            return item.data(index.column())

        if role == QtCore.Qt.UserRole:
            return item

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        node = index.internalPointer()
        if not node:
            return False

        if role == QtCore.Qt.EditRole:
            self.register_edited.emit(index, value)

        return True

class RegistersListView(VQTraceNotifier, vq_tree.VQTreeView):
    '''
    A pure "list view" object for registers
    '''
    def __init__(self, trace=None, parent=None):
        VQTraceNotifier.__init__(self, trace)
        vq_tree.VQTreeView.__init__(self, parent=parent)

        self.descrend = e_mem_rend.AutoBytesRenderer()

        self.setAlternatingRowColors(True)
        # snapped in by someone else.  use a signal instead.
        self.regnames = None
        # used to determine what registers have changed.
        self.lastregs = {}
        self.regvals = {}

        self.setStyleSheet(vq_colors.getDefaultColors())

        model = RegisterListModel(parent=self)
        self.setModel(model)

    def regEdited(self, index, value):
        node = index.internalPointer()
        if not node or not value:
            return False

        expr = str(value.toString())
        try:
            value = self.trace.parseExpression(expr)
        except Exception as e:
            return False

        reg = node.rowdata[0]
        self.trace.setRegisterByName(reg, value)
        regval = self.trace.getRegisterByName(reg)
        # TODO: yuck...implement a "fast" string memorycanvas?
        smc = e_mcanv.StringMemoryCanvas(self.trace)
        self.descrend.render(smc, regval)

        node.rowdata[1] = self.trace.pointerString(regval)
        node.rowdata[2] = regval
        node.rowdata[3] = str(smc)

        return True

    def setModel(self, model):
        model.dataChanged.connect(self.dataChanged)
        model.rowsInserted.connect(self.rowsInserted)
        model.register_edited.connect(self.regEdited)
        return QTreeView.setModel(self, model)

    def vqLoad(self):
        self.lastregs = self.regvals.copy()
        if not self.trace.isAttached():
            self.setEnabled(False)
            return

        if self.trace.isRunning():
            self.setEnabled(False)
            return

        model = RegisterListModel(parent=self)
        self.setModel(model)

        for rname in self.regnames:
            rval = self.trace.getRegisterByName(rname)
            self.regvals[rname] = rval
            hexva = self.trace.pointerString(rval)

            try:
                # TODO: yuck...implement a "fast" string memorycanvas?
                smc = e_mcanv.StringMemoryCanvas(self.trace)
                self.descrend.render(smc, rval)
            except Exception as e:
                smc = repr(e)
            finally:
                model.append((rname, hexva, rval, str(smc)))

class RegColorDelegate(QStyledItemDelegate):

    def __init__(self, parent):
        QStyledItemDelegate.__init__(self, parent)
        self.reglist = parent

    def paint(self, painter, option, index):
        node = index.internalPointer()
        weight = QtGui.QFont.Normal
        if self.reglist.lastregs.get(node.rowdata[0]) != node.rowdata[2]:
            weight = QtGui.QFont.Bold
        option.font.setWeight(weight)
        return QStyledItemDelegate.paint(self, painter, option, index)

class RegistersView(QWidget):
    '''
    A register view which includes the idea of "sub views" for particular
    sets of registers per-architecture.
    '''

    def __init__(self, trace=None, parent=None):
        QWidget.__init__(self, parent=parent)
        self.setWindowTitle('Registers')

        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(2, 2, 2, 2)
        vbox.setSpacing(4)

        self.viewnames = QComboBox(self)
        self.regviews = {}
        self.flagviews = {}

        reg_groups = trace.archGetRegisterGroups()
        for name, group in reg_groups.items():
            self.regviews[name] = group
            self.viewnames.addItem(name)

        self.reglist = RegistersListView(trace=trace, parent=self)
        self.regdelegate = RegColorDelegate(self.reglist)
        self.reglist.setItemDelegate(self.regdelegate)
        # TODO: we should switch theme overall to monospace font
        #font = Qt.QFont('Courier New', 10)
        #self.reglist.setFont(font)

        vbox.addWidget(self.viewnames)

        # show general in dropdown by default if exists, otherwise all
        # (preferences will re-set)
        if 'general' in self.regviews:
            idx = self.viewnames.findText('general')
            self.regViewNameSelected('general')
            self.viewnames.setCurrentIndex(idx)
        else:
            idx = self.viewnames.findText('all')
            self.regViewNameSelected('all')
            self.viewnames.setCurrentIndex(idx)

        statusreg_widget = VQFlagsGridView(trace=trace, parent=self)
        statusreg_widget.setMaximumHeight(60)
        statusreg_widget.hide()

        splitview = QSplitter(QtCore.Qt.Vertical)
        splitview.addWidget(self.reglist)
        splitview.addWidget(statusreg_widget)
        vbox.addWidget(splitview)

        self.viewnames.currentTextChanged.connect(self.regViewNameSelected)

        self.setLayout(vbox)

    def regViewNameSelected(self, name):
        self.reglist.regnames = self.regviews.get(str(name), None)
        self.reglist.vqLoad()

class VQFlagsGridView(VQTraceNotifier, QWidget):
    '''
    Show the state of the status register (if available).
    '''
    def __init__(self, trace=None, parent=None):
        QWidget.__init__(self, parent=parent)
        VQTraceNotifier.__init__(self, trace)

        self.grid = QGridLayout()
        if not trace.hasStatusRegister():
            return

        self.flags_def = trace.getStatusRegNameDesc()
        self.flags = {}
        self.flag_labels = {}
        for idx, (name, desc) in enumerate(self.flags_def):
            flag_button = QPushButton(name)
            flag_button.clicked.connect(self.buttonClicked)
            flag_button.setToolTip(desc)

            flag_label = QLabel('0', self)
            flag_label.setAlignment(QtCore.Qt.AlignCenter)
            self.flag_labels[name] = flag_label

            self.grid.addWidget(flag_button, 0, idx)
            self.grid.addWidget(flag_label, 1, idx)

        self.setLayout(self.grid)
        self.vqLoad()

    def buttonClicked(self):
        obj = self.sender()
        name = obj.text()

        value = self.trace.getRegisterByName(name)
        if value not in (0, 1):
            raise Exception('unexpected value for status register flag')

        value = not bool(value)
        self.trace.setRegisterByName(name, value)

        self.vqLoad()

    def vqLoad(self):
        if not self.trace.isAttached():
            self.setEnabled(False)
            return

        if self.trace.isRunning():
            self.setEnabled(False)
            return

        if not self.trace.hasStatusRegister():
            self.hide()
            self.setEnabled(False)

        self.show()
        self.flags = self.trace.getStatusFlags()
        for name, desc in self.flags_def:
            self.flag_labels[name].setText(str(self.flags[name]))

        self.update()

class VQProcessListModel(vq_tree.VQTreeModel):
    columns = ('Pid', 'Name')

class VQProcessListView(vq_tree.VQTreeView):
    def __init__(self, trace=None, parent=None):
        vq_tree.VQTreeView.__init__(self, parent=parent)
        if trace is None:
            trace = vtrace.getTrace()
        self.trace = trace

        model = VQProcessListModel(parent=self)
        self.setModel(model)
        self.setAlternatingRowColors(True)

        for pid,name in self.trace.ps():
            model.append((pid,name))

class VQProcessSelectDialog(QDialog):

    def __init__(self, trace=None, parent=None):
        QDialog.__init__(self, parent=parent)

        self.pid = None

        self.setWindowTitle('Select a process...')

        vlyt = QVBoxLayout()
        hlyt = QHBoxLayout()

        self.plisttree = VQProcessListView(trace=trace, parent=self)

        hbox = QWidget(parent=self)

        ok = QPushButton("Ok", parent=hbox)
        cancel = QPushButton("Cancel", parent=hbox)

        self.plisttree.doubleClicked.connect( self.dialog_activated )

        ok.clicked.connect(self.dialog_ok)
        cancel.clicked.connect(self.dialog_cancel)

        hlyt.addStretch(1)
        hlyt.addWidget(cancel)
        hlyt.addWidget(ok)
        hbox.setLayout(hlyt)

        vlyt.addWidget(self.plisttree)
        vlyt.addWidget(hbox)
        self.setLayout(vlyt)

        self.resize(300, 500)

    def dialog_activated(self, idx):
        node = idx.internalPointer()
        if node:
            self.pid = node.rowdata[0]
            self.accept()

    def dialog_ok(self):
        for idx in self.plisttree.selectedIndexes():
            node = idx.internalPointer()
            if node:
                self.pid = node.rowdata[0]
                break
        self.accept()

    def dialog_cancel(self):
        self.reject()

@idlethreadsync
def getProcessPid(trace=None, parent=None):
    d = VQProcessSelectDialog(trace=trace, parent=parent)
    r = d.exec_()
    return d.pid

class FileDescModel(vq_tree.VQTreeModel):
    columns = ('Fd','Type','Name')

class VQFileDescView(VQTraceNotifier, vq_tree.VQTreeView):

    def __init__(self, trace, parent=None):
        VQTraceNotifier.__init__(self, trace)
        vq_tree.VQTreeView.__init__(self, parent=parent)
        self.setWindowTitle('File Descriptors')
        self.setModel(FileDescModel(parent=self))
        self.vqLoad()

    def vqLoad(self):

        if not self.trace.isAttached():
            self.setEnabled(False)
            return

        if self.trace.isRunning():
            self.setEnabled(False)
            return

        model = FileDescModel(parent=self)
        for fd,fdtype,bestname in self.trace.getFds():
            model.append((fd, fdtype, bestname))
        self.setModel(model)

class VQTraceToolBar(QToolBar, vtrace.Notifier):

    def __init__(self, trace, parent=None):
        QToolBar.__init__(self, parent=parent)
        vtrace.Notifier.__init__(self)
        self.trace = trace

        self.setObjectName('VtraceToolbar')

        self.attach_action = self.addAction('attach')
        self.attach_action.setToolTip('Attach to a process')
        self.attach_action.triggered.connect(self.actAttach)

        self.detach_action = self.addAction('detach')
        self.detach_action.setToolTip('Detach from current process')
        self.detach_action.triggered.connect(self.actDetach)

        self.continue_action = self.addAction('continue')
        self.continue_action.setToolTip('Continue current process')
        self.continue_action.triggered.connect(self.actContinue)

        self.break_action = self.addAction('break')
        self.break_action.setStatusTip('Break current process')
        self.break_action.triggered.connect(self.actBreak)

        self.stepi_action = self.addAction('stepi')
        self.stepi_action.setToolTip('Single step the current process')
        self.stepi_action.triggered.connect(self.actStepi)

        self.stepover_action = self.addAction('stepover')
        self.stepover_action.setToolTip('step over current instruction')
        self.stepover_action.triggered.connect(self.actStepover)

        trace.registerNotifier(NOTIFY_ALL, self)
        self._updateActions(trace.isAttached(), trace.isRunning())

    def actAttach(self, *args, **kwargs):
        pid = getProcessPid(trace=self.trace)
        if pid is not None:
            workthread(self.trace.attach)(pid)

    @workthread
    def actDetach(self, thing):
        if self.trace.isAttached():
            self.trace.detach()

    @workthread
    def actContinue(self, thing):
        self.trace.run()

    @workthread
    def actBreak(self, thing):
        if self.trace.getMeta('PendingBreak'):
            return
        self.trace.setMeta('PendingBreak', True)
        self.trace.sendBreak()

    @workthread
    def actStepi(self, thing):
        self.trace.stepi()

    def actStepover(self, thing):
        # TODO: move stepover into vtrace instead of vdb?
        raise Exception('no stepover for vtrace. :(')

    @idlethread
    def _updateActions(self, attached, running):
        if not attached:
            self.attach_action.setEnabled(True)
            self.detach_action.setEnabled(False)
            self.continue_action.setEnabled(False)
            self.break_action.setEnabled(False)
            self.stepi_action.setEnabled(False)
            self.stepover_action.setEnabled(False)
        else:
            if running:
                self.attach_action.setEnabled(False)
                self.detach_action.setEnabled(False)
                self.continue_action.setEnabled(False)
                self.break_action.setEnabled(True)
                self.stepi_action.setEnabled(False)
                self.stepover_action.setEnabled(False)
            else:
                self.attach_action.setEnabled(False)
                self.detach_action.setEnabled(True)
                self.continue_action.setEnabled(True)
                self.break_action.setEnabled(True)
                self.stepi_action.setEnabled(True)
                self.stepover_action.setEnabled(True)

    def notify(self, event, trace):
        if event == NOTIFY_DETACH:
            self._updateActions(False, False)

        elif event == NOTIFY_CONTINUE:
            self._updateActions(True, True)

        else:
            self._updateActions(trace.isAttached(), trace.shouldRunAgain())

class VQMemoryMapView(VQTraceNotifier, envi_qt_memmap.VQMemoryMapView):
    '''
    A memory map view which is sensitive to the status of a
    trace object.
    '''
    def __init__(self, trace, parent=None):
        VQTraceNotifier.__init__(self, trace)
        envi_qt_memmap.VQMemoryMapView.__init__(self, trace, parent=parent)

    def vqLoad(self):
        if not self.trace.isAttached():
            self.setEnabled(False)
            return

        if self.trace.isRunning():
            self.setEnabled(False)
            return

        envi_qt_memmap.VQMemoryMapView.vqLoad(self)

class VQThreadListModel(vq_tree.VQTreeModel):
    columns = ('Thread Id','Thread Info', 'State')

class VQThreadsView(vq_tree.VQTreeView, VQTraceNotifier):

    def __init__(self, trace=None, parent=None, selectthread=None):
        # selectthread is an optional endpoint to connect to
        vq_tree.VQTreeView.__init__(self, parent=parent)
        VQTraceNotifier.__init__(self, trace)
        self.setWindowTitle('Threads')
        self.setModel(VQThreadListModel(parent=self))
        self.setAlternatingRowColors(True)
        self.vqLoad()
        self.selectthread = selectthread

    def selectionChanged(self, selected, deselected):
        if len(self.selectedIndexes()) > 0:
            idx = self.selectedIndexes()[0]
            node = idx.internalPointer()
            if node:
                self.trace.selectThread(node.rowdata[0])

            return vq_tree.VQTreeView.selectionChanged(self, selected, deselected)

    def vqLoad(self):

        if not self.trace.isAttached():
            self.setEnabled(False)
            return

        if self.trace.isRunning():
            self.setEnabled(False)
            return

        model = VQThreadListModel(parent=self)

        stid = self.trace.getMeta('ThreadId')
        for i, (tid, tinfo) in enumerate(self.trace.getThreads().items()):
            state = ''
            if self.trace.isThreadSuspended(tid):
                state = 'suspended'
            model.append((tid, tinfo, state))

        self.setModel(model)



"""
Attempt to find locations that are likely vftable arrays
by finding lists of function pointers in memory.

This should be a *late* pass after all code that is possible
is defined.
"""
import logging

import vivisect.const as v_const

logger = logging.getLogger(__name__)


def analyze(vw):

    psize = vw.arch.getPointerSize()

    for lva, pval in vw.findPointers():
        xrto = vw.getXrefsFrom(lva)
        if not xrto:
            continue

        if not vw.isFunction(xrto[0][v_const.XR_TO]):
            continue

        count = 1
        va = lva + psize
        while True:
            if vw.getLocation(va) is not None:
                break

            ptrva = vw.castPointer(va)
            # FIXME this might make us miss stuff
            if not vw.isFunction(ptrva):
                logger.debug("Skipping: 0x%x", ptrva)
                break

            count += 1
            va += psize

        if count >= 4:
            logger.warning("VFTABLE? 0x%.8x", ptrva)



'''
A dynadag-ish graph layout calculator...
'''

import visgraph.layouts as vg_layout
import visgraph.drawing.bezier as vg_bezier

zero_zero = (0,0)

def revenumerate(l):
    return list(zip(range(len(l)-1, -1, -1), reversed(l)))

SCOOCH_LEFT     = 0
SCOOCH_RIGHT    = 1

class DynadagLayout(vg_layout.GraphLayout):

    def __init__(self, graph, barry=10):

        vg_layout.GraphLayout.__init__(self, graph)

        self._addGhostNodes()

        self._barry_count = barry
        self.width_pad = 20
        self.height_pad = 40

    def getLayoutSize(self):
        '''
        Return the width,height of this layout.
        '''
        height = 0
        width  = 0
        for layer in self.layers:

            lheight = 0
            lwidth = 0

            for nid,ninfo in layer:
                xsize, ysize = ninfo.get('size', zero_zero)
                lheight = max(lheight, ysize + self.height_pad)
                lwidth += xsize + self.width_pad

            height += lheight
            width = max(lwidth, width)

        return width, height

    def _baryCenter(self, nid, ninfo):
        tot = 0
        cnt = 0
        for eid, n1, n2, einfo in self.graph.getRefsFromByNid(nid):
            node2 = self.graph.getNode(n2)
            tot += node2[1].get('layerpos')
            cnt += 1

        for eid, n1, n2, einfo in self.graph.getRefsToByNid(nid):
            node1 = self.graph.getNode(n1)
            tot += node1[1].get('layerpos')
            cnt += 1
        barry = 0
        if cnt:
            barry = tot / float(cnt)
        ninfo['barycenter'] = barry

    # Try out "barycenter" averaging and re-ordering.
    def _orderNodesByBary(self):
        # Go through the layers and do barycenter calcs first.
        # FIXME how do we tell when we're done?
        for i in range(self._barry_count):
            for layer in self.layers:
                for nid, ninfo in layer:
                    self._baryCenter(nid, ninfo)

            for layer in self.layers:
                layer.sort(key=lambda k: k[1].get('barycenter'))
                for i, (nid, ninfo) in enumerate(layer):
                    ninfo['layerpos'] = i

    def _getNodeRelPos(self, nid, ninfo):

        weight = ninfo['weight']

        abovepos = []
        for eid, n1, n2, einfo in self.graph.getRefsToByNid(nid):
            fprops = self.graph.getNodeProps(n1)
            if fprops['weight'] != weight-1:
                continue
            abovepos.append(fprops['layerpos'])
        abovepos.sort()

        belowpos = []
        for eid, n1, n2, einfo in self.graph.getRefsFromByNid(nid):
            tprops = self.graph.getNodeProps(n2)
            if tprops['weight'] != weight+1:
                continue
            belowpos.append(tprops['layerpos'])
        belowpos.sort()

        return abovepos, belowpos

    def _getLayerCross(self, layernum):
        ccount = 0

        layer = self.layers[layernum]
        for i in range(1, len(layer)):

            myabove, mybelow = self._getNodeRelPos(*layer[i])
            hisabove, hisbelow = self._getNodeRelPos(*layer[i-1])

            # If I have any nodes above with position lower
            # than any of his, those are cross overs...
            for mya in myabove:
                for hisa in hisabove:
                    if mya < hisa:
                        ccount += 1

            # If I have any nodes below with position lower
            # than any of his, those acre cross overs...
            for myb in mybelow:
                for hisb in hisbelow:
                    if myb < hisb:
                        ccount += 1

        return ccount

    def _bubSortNodes(self):

        # Go through nodes and see if we can re-order children to
        # reduce crossovers...

        for i in range(len(self.layers)):

            layer = self.layers[i]

            reduced = True
            while reduced:

                reduced = False

                # Get the current crossover count for this layer
                score = self._getLayerCross(i)

                # TODO should we do this multipliciative rather than
                # neighbors only?
                for j in range(len(layer)-1):

                    n1 = layer[j]
                    n2 = layer[j+1]

                    layer[j] = n2
                    layer[j+1] = n1

                    newscore = self._getLayerCross(i)
                    # If this was optimal, keep it and continue
                    if newscore < score:
                        reduced = True
                        n1[1]['layerpos'] = j+1
                        n2[1]['layerpos'] = j
                        break

                    # Nope, put it back...
                    layer[j] = n1
                    layer[j+1] = n2
                    
    def _addGhostNodes(self):
        '''
        Translate the hierarchical graph we are given into dynadag
        friendly graph with ghost nodes....
        '''
        weights = self.graph.getHierNodeWeights()

        # First lets take care of any loop edges
        # (These will be nodes in the graph which are marked "reverse=True"
        # but have been added with src/dst swapped to make graphing easier)
        for eid, n1, n2, einfo in self.graph.getEdges():

            if not einfo.get('reverse'):
                continue

            topweight = weights.get(n1)
            botweight = weights.get(n2)

            # In the case of a single block loop, add one ghost and
            # connect them all
            #if n1 == n2:
            if topweight == botweight:
                bridgenode = self.graph.addNode(ghost=True, weight=topweight)
                weights[bridgenode[0]] = topweight

                self.graph.delEdgeByEid(eid)
                self.graph.addEdgeByNids(n1, bridgenode[0], looptop=True)
                self.graph.addEdgeByNids(bridgenode[0], n2, loopbot=True)
                continue

            # For a "reverse" edge, add a node in the weight for each
            # and connect them.
            topnode = self.graph.addNode(ghost=True, weight=topweight)
            weights[topnode[0]] = topweight

            botnode = self.graph.addNode(ghost=True, weight=botweight)
            weights[botnode[0]] = botweight

            self.graph.addEdge(topnode, botnode) # For rendering, these will be normal!

            # Now, remove the "reverse" edge, and add a 'looptop' and 'loopbot' edge
            self.graph.delEdgeByEid(eid)
            self.graph.addEdgeByNids(n1, topnode[0], looptop=True)
            self.graph.addEdgeByNids(botnode[0], n2, loopbot=True)

        # Create ghost nodes for edges which pass through a weight layer
        for eid, n1, n2, einfo in self.graph.getEdges():
            xweight = weights.get(n1, 0)
            yweight = weights.get(n2, 0)
            if xweight + 1 < yweight:
                self.graph.delEdgeByEid(eid)
                while xweight + 1 < yweight:
                    xweight += 1
                    ghostid = self.graph.addNode(ghost=True, weight=xweight)[0]
                    self.graph.addEdgeByNids(n1, ghostid)
                    n1 = ghostid
                self.graph.addEdgeByNids(n1, n2)

    def layoutGraph(self):

        self.maxweight = 0

        for nid, ninfo in self.graph.getNodes():
            self.maxweight = max(ninfo.get('weight', 0), self.maxweight)

        self.layers = [ [] for i in range(self.maxweight + 1) ]

        done = set()
        def doit(node):
            '''
            Roll through all the nodes and assign them positions in their layer (based on weight)
            '''

            if node[0] in done:
                return

            done.add(node[0])
            efrom = self.graph.getRefsFrom(node)

            for eid, n1, n2, einfo in efrom:
                tonode = self.graph.getNode(n2)
                doit(tonode)

            w = node[1].get('weight', 0)

            layer = self.layers[w]
            self.graph.setNodeProp(node, 'layerpos', len(layer))
            layer.append(node)

        # FIXME support more than one root!
        for rootnode in self.graph.getHierRootNodes():
            doit(rootnode)

        # Now lets use positional averaging to order nodes in the layer
        self._orderNodesByBary()
        self._bubSortNodes()

        self.maxwidth = 0

        # Calculate the width / height of each layer...
        lwidths = []        # The width of this total layer
        self.lheights = []       # The tallest node in this layer
        for layer in self.layers:
            x = self.width_pad
            y = 0

            heightmax = 0
            for nid, ninfo in layer:
                size = ninfo.get('size', zero_zero)
                xx, yy = size

                heightmax = max(heightmax, yy)

                x += xx
                y += yy

            lwidths.append(x)
            self.lheights.append(heightmax)

            self.maxwidth = max(self.maxwidth, x)

        # Now that we have them sorted, lets set their individual positions...
        vpad = 0
        for i,layer in enumerate(self.layers):
            hpad = (self.maxwidth - lwidths[i]) / 2
            hpad += self.width_pad
            for nid,ninfo in layer:
                xpos = hpad
                ypos = vpad

                xsize, ysize = ninfo.get('size', zero_zero)

                ninfo['position'] = (xpos,ypos)
                ninfo['vert_pad'] = self.lheights[i] - ysize

                hpad += xsize

                hpad += self.width_pad

            vpad += self.lheights[i]
            vpad += self.height_pad


        # Optimize the positions of nodes by moving them outward to align

        # First from top to bottom
        for i, layer in enumerate(self.layers):

            layermid = len(layer) / 2

            # From the left side, scooch kids out...
            for j, (nid,ninfo) in enumerate(layer):

                if not ninfo.get('ghost'):
                    break

                self._scoochXKids(nid, ninfo, SCOOCH_LEFT)

            # From the right side, scooch kids out...
            for j, (nid, ninfo) in revenumerate(layer):

                if not ninfo.get('ghost'):
                    break

                self._scoochXKids(nid, ninfo, SCOOCH_RIGHT)

        # From the bottom to the top!
        for i, layer in revenumerate(self.layers):

            layermid = len(layer) / 2

            # From the left side, scooch kids out...
            for j, (nid,ninfo) in enumerate(layer):

                if not ninfo.get('ghost'):
                    break

                self._scoochXParents(nid, ninfo, SCOOCH_LEFT)

            # From the right side, scooch kids out...
            for j, (nid, ninfo) in revenumerate(layer):

                if not ninfo.get('ghost'):
                    break

                self._scoochXParents(nid, ninfo, SCOOCH_RIGHT)

        # Finally, we calculate the drawing for the edge lines
        self._calcEdgeLines()

    def _scoochXParents(self, nid, ninfo, lr=None):

        weight = ninfo['weight']

        for eid, n1, n2, einfo in self.graph.getRefsToByNid(nid):

            pinfo = self.graph.getNodeProps(n1)

            # Only do ghost nodes (for now...)
            if not pinfo.get('ghost'):
                continue

            # Only do this to parents in the layer above us...
            if pinfo['weight'] != weight-1:
                continue

            self._scoochXAlign(ninfo, pinfo, lr=lr)

    def _scoochXKids(self, nid, ninfo, lr=None):

        weight = ninfo['weight']

        for eid, n1, n2, einfo in self.graph.getRefsFromByNid(nid):

            kinfo = self.graph.getNodeProps(n2)

            # Only do ghost nodes (for now...)
            if not kinfo.get('ghost'):
                continue

            # Only do this to kids in the layer beneath us...
            if kinfo['weight'] != weight+1:
                continue

            self._scoochXAlign(ninfo, kinfo, lr=lr)

    def _scoochXAlign(self, ninfo, kinfo, lr=None):
        '''
        If possible, move the "kidinfo" node toward ninfo
        along the X axis...  If "lr" is specified, only move
        the "kidnode" (which may be "above" you...) if it is
        moving either SCOOCH_LEFT or SCOOCH_RIGHT as specified.
        '''
        xpos, ypos = ninfo['position']
        xsize, ysize = ninfo.get('size', zero_zero)
        xmid = xpos + ( xsize / 2 )

        kxpos, kypos = kinfo['position']
        kxsize, kysize = kinfo.get('size', zero_zero)
        kxmid = kxpos + ( kxsize / 2 )

        xdelta = xmid - kxmid

        # If they only want us to go left, and the delta
        # is right, bail...
        if lr == SCOOCH_LEFT and xdelta >= 0:
            return

        # If they only want us to go right, and the delta
        # is left, bail...
        if lr == SCOOCH_RIGHT and xdelta <= 0:
            return

        self._scoochX(kinfo, xdelta)

    def _scoochX(self, ninfo, xdelta):

        layerpos = ninfo.get('layerpos')
        x, y = ninfo['position']
        xsize, ysize = ninfo.get('size', zero_zero)
        layer = self.layers[ninfo['weight']]
        layermax = len(layer) - 1

        # There's always room on the left if we're the first...
        if layerpos == 0 and xdelta < 0:
            ninfo['position'] = (x+xdelta, y)
            return

        # Always room on the right if we're last!
        if layerpos == layermax and xdelta > 0:
            ninfo['position'] = (x+xdelta, y)
            return

        # Sigh... now we have to get fancy...

        # If they're asking us to go left, find out about our
        # left sibling
        if xdelta < 0:
            snid, sinfo = layer[layerpos - 1]
            sx, sy = sinfo['position']
            sxsize, sysize = sinfo.get('size', zero_zero)

            sright = (sx + sxsize) + self.width_pad

            #leftroom = sright - x
            # "greater" is less movement here...
            xdelta = max(xdelta, sright - x)
            ninfo['position'] = (x+xdelta, y)
            return

        # If they're asking us to go right, find out about our
        # right sibling
        if xdelta > 0:
            snid, sinfo = layer[layerpos + 1]
            sx, sy = sinfo['position']
            sxsize, sysize = sinfo.get('size', zero_zero)

            myright = x + xsize + self.width_pad

            xdelta = min(xdelta, sx-myright)
            ninfo['position'] = (x+xdelta, y)
            return

    def _calcEdgeLines(self):

        h_hpad = self.width_pad / 2
        h_vpad = self.height_pad / 2

        for eid, n1, n2, einfo in self.graph.getEdges():

            pre_lines = []
            post_lines = []

            pinfo = self.graph.getNodeProps(n1)
            kinfo = self.graph.getNodeProps(n2)

            pwidth, pheight = pinfo.get('size', (0,0))
            pweight = pinfo.get('weight')
            lheight = self.lheights[pweight]
            voffset = lheight - pheight

            if einfo.get('looptop'):

                x1, y1 = vg_layout.entry_pos(pinfo)
                x2, y2 = vg_layout.entry_pos(kinfo)

                xhalf = (x1 - x2) / 2

                b = [ (x1, y1),
                      (x1, y1 - h_vpad),
                      (x2, y2 - h_vpad),
                      (x2, y2),
                    ]

            elif einfo.get('loopbot'):

                x1, y1 = vg_layout.exit_pos(pinfo)
                x2, y2 = vg_layout.exit_pos(kinfo)

                kwidth, kheight = kinfo.get('size', (0,0))
                kweight = kinfo.get('weight')
                klheight = self.lheights[kweight]

                kvoffset = klheight - kheight

                pre_lines = [(x1, y1), (x1, y1 + voffset)]
                post_lines = [(x2, y2), (x2, y2 + kvoffset)]

                b = [ (x1, y1 + voffset),
                      (x1, y1 + voffset + h_vpad),
                      (x2, y2 + kvoffset + h_vpad),
                      (x2, y2 + kvoffset),
                    ]

            else:

                x1, y1 = vg_layout.exit_pos(pinfo)
                x2, y2 = vg_layout.entry_pos(kinfo)

                pre_lines = [(x1,y1), (x1, y1 + voffset)]

                b = [ (x1, y1 + voffset),
                      (x1, y1 + voffset + h_vpad),
                      (x2, y2 - h_vpad),
                      (x2, y2),
                    ]

            bez_lines = vg_bezier.calculate_bezier(b, 20)

            einfo['edge_points'] = pre_lines + bez_lines + post_lines
            #einfo['edge_points'] = bez_lines




import optparse

import envi.cli as e_cli
import vdb.extensions.windows as v_ext_windows
import vtrace.platforms.winkern as vt_winkern
# FIXME bugcheck lookup

KERNEL_MODE_EXCEPTION_NOT_HANDLED = 0x0000008e

def getBugCheckName(code):
    return 'FIXME BUGCHECK NAME'

'''
//
// Flags used in the VAD
//
typedef struct _MMVAD_FLAGS
{
    ULONG CommitCharge:19;
    ULONG NoChange:1;
    ULONG VadType:3;
    ULONG MemCommit:1;
    ULONG Protection:5;
    ULONG Spare:2;
    ULONG PrivateMemory:1;
} MMVAD_FLAGS, *PMMVAD_FLAGS;

//
// Extended flags used in the VAD
//
typedef struct _MMVAD_FLAGS2
{
    ULONG FileOffset:24;
    ULONG SecNoChange:1;
    ULONG OneSecured:1;
    ULONG MultipleSecured:1;
    ULONG ReadOnly:1;
    ULONG LongVad:1;
    ULONG ExtendableFile:1;
    ULONG Inherit:1;
    ULONG CopyOnWrite:1;
} MMVAD_FLAGS2, *PMMVAD_FLAGS2;
'''

#b1971ba0 (140) KTRAP_FRAME: KTRAP_FRAME
#b1971ba0 (04)   DbgEbp: 0xb1971c1c (2979470364)
#b1971ba4 (04)   DbgEip: 0x80536e9e (2152951454)
#b1971ba8 (04)   DbgArgMark: 0xbadb0d00 (3134917888)
#b1971bac (04)   DbgArgPointer: 0x00000002 (2)
#b1971bb0 (04)   TempSegCs: 0xe1371848 (3778484296)
#b1971bb4 (04)   TempEsp: 0x00000000 (0)
#b1971bb8 (04)   Dr0: 0x8212f810 (2182281232)
#b1971bbc (04)   Dr1: 0x00000000 (0)
#b1971bc0 (04)   Dr2: 0x00000000 (0)
#b1971bc4 (04)   Dr3: 0xb1971c24 (2979470372)
#b1971bc8 (04)   Dr6: 0x8055b740 (2153101120)
#b1971bcc (04)   Dr7: 0x805453e1 (2153010145)
#b1971bd0 (04)   SegGs: 0x00000000 (0)
#b1971bd4 (04)   SegEs: 0x00000023 (35)
#b1971bd8 (04)   SegDs: 0x00000023 (35)
#b1971bdc (04)   Edx: 0x00000002 (2)
#b1971be0 (04)   Ecx: 0x00000000 (0)
#b1971be4 (04)   Eax: 0xb1971c5c (2979470428)
#b1971be8 (04)   PreviousPreviousMode: 0xe1371848 (3778484296)
#b1971bec (04)   ExceptionList: 0xb1971c80 (2979470464)
#b1971bf0 (04)   SegFs: 0x00000030 (48)
#b1971bf4 (04)   Edi: 0x00000000 (0)
#b1971bf8 (04)   Esi: 0xb1971c64 (2979470436)
#b1971bfc (04)   Ebx: 0x00000000 (0)
#b1971c00 (04)   Ebp: 0xb1971c1c (2979470364)
#b1971c04 (04)   ErrCode: 0x00000002 (2)
#b1971c08 (04)   Eip: 0x80536e9e (2152951454)
#b1971c0c (04)   SegCs: 0x00000008 (8)
#b1971c10 (04)   EFlags: 0x00010293 (66195)
#b1971c14 (04)   HardwareEsp: 0x82331648 (2184386120)
#b1971c18 (04)   HardwareSegSs: 0x00000002 (2)
#b1971c1c (04)   V86Es: 0xb1971c3c (2979470396)
#b1971c20 (04)   V86Ds: 0x8052819f (2152890783)
#b1971c24 (04)   V86Fs: 0x00000000 (0)
#b1971c28 (04)   V86Gs: 0xb1971c64 (2979470436)

def bugcheck(db, line):
    '''
    Analysze and print out the current state of a windows
    bugcheck ( blue screen ).

    Usage: bugcheck
    '''
    trace = db.getTrace()
    sp = trace.getStackCounter()
    pc = trace.getProgramCounter()
    sym = trace.getSymByAddr(pc)
    if sym is None:
        db.vprint('Not currently at known bugcheck entry: 0x%.8x' % pc)
        return

    db.vprint('')

    sig = trace.getMeta('PendingSignal')
    symname = str(sym)

    if symname == 'nt.KeBugCheck':
        db.vprint('KeBugCheck( 0x%.8x ) - %s' % (sig, getBugCheckName(sig)))
        db.vprint('(no further info available from KeBugCheck)')
        return

    if symname == 'nt.KeBugCheckEx':
        db.vprint('KeBugCheckEx( 0x%.8x ) - %s' % (sig, getBugCheckName(sig)))
        db.vprint('')
        # FIXME use calling convention / impapi?
        if sig == KERNEL_MODE_EXCEPTION_NOT_HANDLED:
            retaddr,stacksig,excode,exaddr,extrap = trace.readMemoryFormat(sp, '<5I')

            trap = trace.getStruct('nt.KTRAP_FRAME', extrap)
            #db.vprint( trap.tree(va=extrap) )
            #db.vprint('')
            segs1 = ( trap.SegDs, trap.V86Ds, trap.SegEs, trap.V86Es )
            segs2 = ( trap.SegFs, trap.V86Fs, trap.SegGs, trap.V86Gs )
            db.vprint('Segments:')
            db.vprint('ds=%.2d:0x%.8x es=%.2d:0x%.8x' % segs1 )
            db.vprint('fs=%.2d:0x%.8x gs=%.2d:0x%.8x' % segs2 )

            db.vprint('')
            db.vprint('Registers:')
            db.vprint('eax=0x%.8x ebx=0x%.8x ecx=0x%.8x edx=0x%.8x' % (trap.Eax,trap.Ebx,trap.Ecx,trap.Edx))
            db.vprint('esi=0x%.8x edi=0x%.8x ebp=0x%.8x esp=0x????????' % (trap.Esi,trap.Edi,trap.Ebp))

            db.vprint('')

            db.vprint('Exception Trace:')

            eip = exaddr
            ebp = trap.Ebp

            while True:

                args = []
                bpvalid = trace.isValidPointer(ebp)
                if bpvalid:
                    args = trace.readMemoryFormat(ebp + 8, '<6I')

                loopbp = ebp

                argstr = ' '.join([ ('0x%.8x' % a) for a in args ])

                framestr = '0x%.8x [0x%.8x] %s' % (eip,ebp,argstr)

                sym = trace.getSymByAddr(eip, exact=False)
                if sym is not None:
                    framestr += ' (%s + %d)' % (sym, eip - int(sym))

                db.vprint(framestr)

                if not bpvalid:
                    break

                ebp,eip = trace.readMemoryFormat(ebp, '<2I')

                if ebp == loopbp:
                    break

            sym = trace.getSymByAddr(exaddr, exact=False)
            db.vprint('')
            db.vprint('Exception Trap: 0x%.8x' % extrap)
            db.vprint('Exception Addr: 0x%.8x (%s + %d)' % (exaddr, sym, exaddr-int(sym)))
            db.vprint('Exception Code: 0x%.8x' % excode)
            db.vprint('Exception Inst: %r' % trace.parseOpcode(exaddr))

def ktrap(db, line):
    '''
    Parse a KTRAP_FRAME structure from the given memory address.

    Usage: ktrap <addr_expression>
    '''
    trace = db.getTrace()
    trapaddr = trace.parseExpression( line )
    trap = trace.getStruct('nt.KTRAP_FRAME', trapaddr)
    db.vprint( trap.tree(va=trapaddr) )

def ethread(db, line):
    '''
    Display the details for the ethreads.

    Usage: ethread [addrexpr]
    '''
    t = db.getTrace()
    kpcr = t.getStruct('nt.KPCR', va=t.getVariable('kpcr'))
    if not line:
        db.vprint('CurrentThread: 0x%.8x' % kpcr.PrcbData.CurrentThread)
        db.vprint('NextThread: 0x%.8x' % kpcr.PrcbData.CurrentThread)
        return
    va = t.parseExpression(line)
    ethread = t.getStruct('nt.ETHREAD',va=va)
    db.vprint(ethread.tree(va=va))

def vads(db,line):
    '''
    Walk and display the VADs from the specified root.

    Usage: vads <addrexpr>
    '''
    t = db.getTrace()
    if not line:
        return db.do_help('vads')

    vadroot = t.parseExpression(line)

    vads = list(walkVads(db,t,vadroot))
    vads.sort(cmp=lambda x,y: cmp(x[1].StartingVpn,y[1].StartingVpn))

    for va,vad in vads:
        start = vad.StartingVpn << t.getMeta('PAGE_SHIFT')
        end = vad.EndingVpn << t.getMeta('PAGE_SHIFT')
        flags = vad.u.LongFlags
        db.vprint('vad: 0x%.8x 0x%.8x-0x%.8x 0x%.8x' % (va,start,end,flags))

def walkVads(db,trace,vadroot):
    todo = [ (vadroot,trace.getStruct('nt.MMVAD_SHORT',va=vadroot)), ]
    while todo:
        va,vad = todo.pop()
        yield va,vad

        l = vad.LeftChild
        r = vad.RightChild

        if l: todo.append((l,trace.getStruct('nt.MMVAD_SHORT',va=l)))
        if r: todo.append((r,trace.getStruct('nt.MMVAD_SHORT',va=r)))

def walkEprocesses(db,trace):
    dbgdata64 = db.getRunCacheVar('KDDEBUGGER_DATA64')
    phead = dbgdata64.PsActiveProcessHead
    for va,obj in vt_winkern.walkListEntryHead(trace,phead,'nt.EPROCESS','ActiveProcessLinks'):
        yield va,obj

def walkEThreads(db,trace,listva):
    for va,obj in vt_winkern.walkListEntryHead(trace,listva,'nt.ETHREAD','ThreadListEntry'):
        yield va,obj

def walkHandleTables(db,trace,listva):
    for va,obj in vt_winkern.walkListEntry(trace,listva,'nt.HANDLE_TABLE','HandleTableList'):
        yield va,obj

def showeproc(db,trace,opts,eprocva,eproc):
    pname = ''.join([ chr(c[1]) for c in eproc.ImageFileName ]).strip('\x00')

    peb = eproc.Peb
    vadroot = eproc.VadRoot
    objtable = eproc.ObjectTable
    db.vprint('eprocess: 0x%.8x vad: 0x%.8x peb: 0x%.8x objs: 0x%.8x %s' % (eprocva, vadroot, peb, objtable, pname))

    if opts.dothreads:
        thead = eprocva + eproc.vsGetOffset('ThreadListHead')
        for eva,ethread in walkEThreads(db,trace,thead):
            db.vprint('  ethread: 0x%.8x' % eva)

    if opts.dohandles:
        thead = eprocva + eproc.vsGetOffset('ThreadListHead')
        for hva,htable in walkEThreads(db,trace,thead):
            db.vprint('  handle_table: 0x%.8x' % hva)

def eproc(db, line):
    '''
    Display the details for the eprocesses.

    Usage: eproc [addrexpr]
    '''
    argv = e_cli.splitargs(line)

    parser = optparse.OptionParser('eproc')
    parser.add_option('--name',dest='doname',default=None)
    parser.add_option('--threads',dest='dothreads',default=False,action='store_true')
    parser.add_option('--handles',dest='dohandles',default=False,action='store_true')
    parser.add_option('--objects',dest='doobjects',default=False,action='store_true')

    opts,argv = parser.parse_args(argv)

    t = db.getTrace()

    if argv:
        for expr in argv:
            eprocva = t.parseExpression(expr)
            eproc = t.getStruct('nt.EPROCESS',eprocva)
            showeproc(db,t,opts,eprocva,eproc)
        return

    dbgdata64 = db.getRunCacheVar('KDDEBUGGER_DATA64')

    phead = dbgdata64.PsActiveProcessHead
    for eprocva,eproc in walkEprocesses(db,t):
        if opts.doname:
            pname = ''.join([ chr(c[1]) for c in eproc.ImageFileName ]).strip('\x00').lower()
            if opts.doname.lower() != pname:
                continue

        showeproc(db,t,opts,eprocva,eproc)
        #pname = ''.join([ chr(c[1]) for c in eproc.ImageFileName ]).strip('\x00')
        #db.vprint('eprocess: 0x%.8x %s' % (eprocva, pname))
        #thead = eprocva + eproc.vsGetOffset('ThreadListHead')
        #for eva,ethread in walkEThreads(db,t,thead):
            #db.vprint('  ethread: 0x%.8x' % eva)

        #if pname == 'calc.exe':
            #mynext,myprev = t.readMemoryFormat(elist,'<II')
            #t.writeMemoryFormat(myprev, '<I', mynext)
            #t.writeMemoryFormat(mynext+4, '<I', myprev)

        #elist = t.readMemoryFormat(elist,'<I')[0]


    #kpcr = t.getStruct('nt.KPCR', va=t.getVariable('kpcr'))
    #ethread = t.getStruct('nt.ETHREAD', va=kpcr.PrcbData.CurrentThread)
    #print ethread.tree(va=kpcr.PrcbData.CurrentThread)
    #eprocfirst = ethread.ThreadsProcess
    #db.vprint('EPROC: 0x%.8x' % eprocfirst)
    #eproc = t.getStruct('nt.EPROCESS', va=eprocfirst)
    #print eproc.tree()
    #print eproc.ImageFileName
    #eproc.Pcb.
    #listoff = eproc.vsGetOffset(

    #va = t.parseExpression(line)
    #ethread = t.getStruct('nt.ETHREAD',va=va)
    #db.vprint(ethread.tree(va=va))

def pools(db,line):
    '''
    Display various information about the kernel allocation
    pools. ( Default lists the non-paged pools )

    Usage: pools [options]
    '''
    t = db.getTrace()

    dbgdata64va = t.getVariable('kddebuggerdata64')
    dbgdata64 = vt_winkern.KDDEBUGGER_DATA64()
    dbgdata64.vsParse(t.readMemory(dbgdata64va, len(dbgdata64)))

    s = dbgdata64.MmNonPagedPoolStart
    e = dbgdata64.MmNonPagedPoolEnd
    db.vprint('Non-Paged Pool: 0x%.8x - 0x%.8x (%d bytes)' % (s,e,(e-s)))
    nppool = dbgdata64.NonPagedPoolDescriptor
    pooldesc = t.getStruct('nt.POOL_DESCRIPTOR',va=nppool)
    #print(pooldesc.tree(va=nppool))

def _ctor_KDDEBUGGER_DATA64(db):
    t = db.getTrace()
    dbgdata64va = t.getVariable('kddebuggerdata64')
    dbgdata64 = vt_winkern.KDDEBUGGER_DATA64()
    dbgdata64.vsParse(t.readMemory(dbgdata64va, len(dbgdata64)))
    return dbgdata64

def vdbExtension(db, trace):
    db.registerCmdExtension(vads,subsys='winkern')
    db.registerCmdExtension(eproc,subsys='winkern')
    db.registerCmdExtension(ktrap,subsys='winkern')
    db.registerCmdExtension(pools,subsys='winkern')
    db.registerCmdExtension(bugcheck,subsys='winkern')
    db.registerCmdExtension(ethread,subsys='winkern')
    db.registerCmdExtension(v_ext_windows.pe,subsys='winkern')

    db.addRunCacheCtor('KDDEBUGGER_DATA64',_ctor_KDDEBUGGER_DATA64)

#bugcheck(db,'')



'''
Various codeflow oriented graph constructs.
'''
import envi
import visgraph.graphcore as v_graphcore

from vivisect.const import *


class CallGraph(v_graphcore.HierGraph):
    '''
    A graph which represents procedural branches.
    '''
    def __init__(self):
        super().__init__()

    def getFunctionNode(self, va):
        node = self.getNode(va)
        if node is None:
            node = self.addNode(nid=va)
        return node

    def getCallEdge(self, f1va, f2va):
        f1 = self.getFunctionNode(f1va)

        # deconflict call graph edges...
        for edge in self.getRefsFrom(f1):
            if edge[2] == f2va:
                return edge

        f2 = self.getFunctionNode(f2va)
        return self.addEdge(f1, f2)


class CodeBlockGraph(v_graphcore.HierGraph):

    def __init__(self, vw):
        super().__init__()
        self.vw = vw
        self.nodevas = {}

    def addEntryPoint(self, va):
        node = self.getNode(va)
        if node is not None:
            return node

        # entry point, by de-facto has a node
        enode = self.getCodeBlockNode(va)

        done = set()

        todo = [va,]
        while todo:

            va = todo.pop()
            if va in done:
                continue

            done.add(va)

            branches = self._getCodeBranches(va)
            tdone = set()
            for tova, bflags in branches:
                if tova in tdone:
                    continue

                tdone.add(tova)
                node = self.getNodeByVa(va)
                if self._addCodeBranch(node, va, tova, bflags):
                    todo.append(tova)

        return enode

    def _getCodeBranches(self, va):
        loc = self.vw.getLocation(va)
        if loc is None or loc[L_LTYPE] != LOC_OP:
            return []

        lva, lsize, ltype, ltinfo = loc

        xrefs = self.vw.getXrefsFrom(va, rtype=REF_CODE)

        crefs = [(xto,xflags) for (xfrom,xto,xtype,xflags) in xrefs]

        # If any of our other branches are conditional, so is our fall
        if not ltinfo & envi.IF_NOFALL:

            bflags = envi.BR_FALL
            if any([(x[3] & envi.BR_COND) for x in xrefs]):
                bflags |= envi.BR_COND

            crefs.append((lva+lsize, bflags))

        return crefs

    def _addCodeBranch(self, node, va, brva, bflags):
        if self.isCodeBlockNode(brva):
            self.addCodeBlockEdge(node, va, brva)
            return True

        if bflags & envi.BR_FALL and not bflags & envi.BR_COND:
            self.addVaToNode(node, brva)
            return True

        if bflags & envi.BR_DEREF:
            # FIXME handle these
            return False

        n2node = self.addCodeBlockEdge(node, va, brva)

        if bflags & envi.BR_PROC:
            self.setNodeProp(n2node, 'isfunc', True)

        return True

    def isCodeBlockNode(self, va):
        return self.getNode(va) is not None

    def getCodeBlockBounds(self, node):
        cbva = node[0]
        lastva = node[1]['valist'][-1]
        cbsize = (lastva - cbva) + 1
        return cbva, cbsize

    def getCodeBlockNode(self, va):
        '''
        Create or retrieve a codeblock node for the given va.

        NOTE: If the given va is already present within another
        node, this API will *split* the other node.
        '''
        # is it already a cb node?
        node = self.getNode(va)
        if node is not None:
            return node

        # is it part of another block already?
        node = self.getNodeByVa(va)
        newnode = self.addNode(nid=va, cbva=va, valist=())
        self.addVaToNode(newnode, va)

        if node is None:
            return newnode

        # we need to split an existing node... neato...
        valist = node[1]['valist']
        vaidx = valist.index(va)

        vabeg = valist[:vaidx]
        vaend = valist[vaidx:]

        lastva = vabeg[-1]

        self.setNodeVaList(node, vabeg)
        self.setNodeVaList(newnode, vaend)

        # steal all his outbound codeflow edges
        for edge in self.getRefsFrom(node):
            codeflow = edge[3].get('codeflow')
            if codeflow is None:
                continue
            self.addCodeBlockEdge(newnode, codeflow[0], codeflow[1])
            self.delEdge(edge)

        # add an outbound to us...
        self.addCodeBlockEdge(node, lastva, va)
        return newnode

    def addCodeBlockEdge(self, node1, va1, va2):
        vatup = (va1, va2)

        edges = self.getEdgesByProp('codeflow', vatup)
        if len(edges):
            return edges[0]

        node2 = self.getCodeBlockNode(va2)
        edge = self.addEdge(node1, node2)

        self.setEdgeProp(edge, 'va1', va1)
        self.setEdgeProp(edge, 'va2', va2)
        self.setEdgeProp(edge, 'codeflow', vatup)

        #w1 = node1[1].get('weight',0)
        #w2 = node2[1].get('weight',0)
        # track weights in real time ( per func? )
        #self.setNodeProp(node2,'weight',max(w2,w1+1))

        return node2

    def addVaToNode(self, node, va):
        self.nodevas[va] = node
        valist = node[1]['valist']
        self.setNodeProp(node, 'valist', valist + (va,))

    def setNodeVaList(self, node, valist):
        [self.nodevas.pop(va,None) for va in node[1]['valist']]
        [self.nodevas.__setitem__(va,node) for va in valist]
        self.setNodeProp(node, 'valist', valist)

    def getNodeByVa(self, va):
        return self.nodevas.get(va)


class FuncBlockGraph(CodeBlockGraph):

    def __init__(self, vw, fva):
        CodeBlockGraph.__init__(self, vw)
        root = self.addEntryPoint(fva)
        self.setHierRootNode(root)

    def _getCodeBranches(self, va):
        return [x for x in CodeBlockGraph._getCodeBranches(self,va) if not x[1] & envi.BR_PROC]



'''
Connect any Exports we have to Imports we may also have 
(most useful for multiple file workspaces)
'''
import logging
logger = logging.getLogger(__name__)

from vivisect import LOC_OP, REF_CODE

def analyze(vw):
    """
    Look for any "imported" symbols that are satisfied by current exports.
    Wire up the connection.

    Currently this is simply a pointer write at the location of the Import.
    If this behavior is ever insufficient, we'll want to track the special
    nature through the Export/Import events.
    """
    logger.info('linking Imports with Exports')
    # store old setting and set _supervisor mode (so we can write wherever
    # we want, regardless of permissions)
    with vw.getAdminRights():
        for iva, isz, itype, isym in vw.getImports():
            impfname, impsym = isym.split('.', 1)
            for eva, num, esym, efname in vw.getExports():
                if impsym != esym:
                    continue

                if impfname not in (efname, '*'):
                    continue

                if vw.isFunctionThunk(eva):
                    logger.info("Skipping Exported Thunk")
                    continue

                # file and symbol name match.  apply the magic.
                # do we ever *not* write in the full address at the import site?
                logger.debug("connecting Import 0x%x -> Export 0x%x (%r)", iva, eva, isym)
                vw.writeMemoryPtr(iva, eva)

                # remove the LOC_IMPORT and make it a Pointer instead
                vw.delLocation(iva)
                vw.makePointer(iva, follow=False) # don't follow, it'll be analyzed later?

                # store the former Import in a VaSet
                vw.setVaSetRow('ResolvedImports', (iva, isym, eva))

                # check if any xrefs to the import are branches and make code-xrefs for them
                for xrfr, xrto, xrt, xrflags in vw.getXrefsTo(iva):
                    loc = vw.getLocation(xrfr)
                    if not loc:
                        continue

                    lva, lsz, ltype, ltinfo = loc
                    if ltype != LOC_OP:
                        logger.warning("XREF not from an Opcode: 0x%x -> 0x%x  (%r)", lva, eva, loc)
                        continue

                    vw.addXref(lva, eva, REF_CODE)
                    logger.debug("addXref(0x%x -> 0x%x)", lva, eva)




import unittest

import vstruct
import vstruct.primitives as p

class NNestedStruct(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)

        self.alpha = p.v_uint64()
        self.beta = vstruct.VStruct()

class NestedStruct(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)

        self.foo = p.v_bytes(size=3)
        self.bar = p.v_uint32()
        self.baz = p.v_bytes(size=256)
        self.faz = NNestedStruct()

class TestStruct(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)

        self.one = p.v_uint32()
        self.two = NestedStruct()
        self.three = p.v_uint32()
        self.four = p.v_bytes(size=100)

class VStructTests(unittest.TestCase):
    def setUp(self):
        self.s = TestStruct()

    def test_vsGetFieldByOffset_start(self):
        fname, fobj = self.s.vsGetFieldByOffset(0)
        self.assertEqual(fname, 'one')
        self.assertIs(fobj, self.s.vsGetField('one'))

    def test_vsGetFieldByOffset_end(self):
        fname, fobj = self.s.vsGetFieldByOffset(len(self.s) - 1)
        self.assertEqual(fname, 'four')
        self.assertIs(fobj, self.s.vsGetField('four'))

    def test_vsGetFieldByOffset_maxval_neg(self):
        with self.assertRaisesRegex(Exception, 'Invalid Offset Specified'):
            tup = self.s.vsGetFieldByOffset(-1)

    def test_vsGetFieldByOffset_maxval_pos(self):
        with self.assertRaisesRegex(Exception, 'Invalid Offset Specified'):
            tup = self.s.vsGetFieldByOffset(0xffffffff)

    def test_vsGetFieldByOffset_maxval_plus1_pos(self):
        with self.assertRaisesRegex(Exception, 'Invalid Offset Specified'):
            tup = self.s.vsGetFieldByOffset(len(self.s))

    def test_vsGetFieldByOffset_nested1(self):
        fname, fobj = self.s.vsGetFieldByOffset(5)
        self.assertEqual(fname, 'two.foo')
        self.assertIs(fobj, self.s.two.vsGetField('foo'))

    def test_vsGetFieldByOffset_nested2(self):
        fname, fobj = self.s.vsGetFieldByOffset(8)
        self.assertEqual(fname, 'two.bar')
        self.assertIs(fobj, self.s.two.vsGetField('bar'))

    def test_vsGetFieldByOffset_nested3(self):
        fname, fobj = self.s.vsGetFieldByOffset(12)
        self.assertEqual(fname, 'two.baz')
        self.assertIs(fobj, self.s.two.vsGetField('baz'))

    def test_vsGetFieldByOffset_nested4(self):
        fname, fobj = self.s.vsGetFieldByOffset(267)
        self.assertEqual(fname, 'two.faz.alpha')
        self.assertIs(fobj, self.s.two.faz.vsGetField('alpha'))

# TODO: could use envi.bits, but do we really want envi dep by default?
blkup = {}
bwidths = (8, 16, 24, 32, 64, )
for bwidth in bwidths:
    umax = (2**bwidth) - 1
    umin = 0

    smax = (2**(bwidth-1)) - 1
    smin = -(2**(bwidth-1))

    blkup[bwidth] = (umin, umax, smin, smax)

class IntegerStruct(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)

class VStructTypeTests(unittest.TestCase):
    def setUp(self):
        self.s = IntegerStruct()

def getTestFunc(name, vsval, val, expval):

    def func(self):
        self.s.vsAddField(name, vsval)

        setattr(self.s, name, val)

        rval = getattr(self.s, name)
        self.assertEqual(rval, expval)

    return func

tdefs = []

# dynamically generate the test definitions
# width, pname, vtype, test value, expected value
for bwidth in bwidths:
    umin, umax, smin, smax = blkup[bwidth]

    for ttype, mmin, mmax in ( ('u', umin, umax), ('', smin, smax), ):
        pname = '{}int{}'.format(ttype, bwidth)
        vtype = 'v_{}int{}'.format(ttype, bwidth)
        vtype = getattr(p, vtype)

        tup = (bwidth, pname, vtype, mmin, mmin)
        tdefs.append(tup)

        tup = (bwidth, pname, vtype, mmax, mmax)
        tdefs.append(tup)

        tup = (bwidth, pname, vtype, mmin-1, mmax)
        tdefs.append(tup)

        tup = (bwidth, pname, vtype, mmin-2, mmax-1)
        tdefs.append(tup)

        tup = (bwidth, pname, vtype, mmax+1, mmin)
        tdefs.append(tup)

        tup = (bwidth, pname, vtype, mmax+2, mmin+1)
        tdefs.append(tup)

# generate unittest functions based on the test definitions
for width, pname, vtype, val, expval in tdefs:
    tfunc = getTestFunc(pname, vtype(), val, expval)
    tname = 'test_{}_{}_{}'.format(pname, val, expval)
    setattr(VStructTypeTests, tname, tfunc)



'''
A package to contain all the extended functionality for platform specific
commands and modules.
'''

import os
import sys
import logging
import importlib
import traceback

logger = logging.getLogger(__name__)

__all__ = ['loadExtensions', 'windows', 'i386', 'darwin', 'amd64', 'gdbstub', 'arm', 'android', 'winkern']


def loadExtensions(vdb, trace):
    '''
    Actually load all known extensions here.
    '''
    plat = trace.getMeta('Platform').lower()
    arch = trace.getMeta('Architecture').lower()

    if plat in __all__:
        mod = __import__('vdb.extensions.%s' % plat, 0, 0, 1)
        mod.vdbExtension(vdb, trace)

    if arch in __all__:
        mod = __import__('vdb.extensions.%s' % arch, 0, 0, 1)
        mod.vdbExtension(vdb, trace)

    extdir = os.getenv('VDB_EXT_PATH')
    if extdir is None:
        extdir = os.path.abspath(os.path.join('vdb', 'ext'))

    for dirname in extdir.split(os.pathsep):

        if not os.path.isdir(dirname):
            vdb.vprint('Invalid VDB_EXT_PATH dir: %s' % dirname)
            continue

        # allow extensions to import from the extensions directory.
        if dirname not in sys.path:
            sys.path.append(dirname)

        for fname in os.listdir(dirname):
            modpath = os.path.join(dirname, fname)
            if os.path.isdir(modpath):
                modpath = os.path.join(modpath, '__init__.py')
                if not os.path.exists(modpath):
                    continue

            if not fname.endswith('.py') or fname == '__init__.py':
                continue

            try:
                # Build code objects from the module files
                spec = importlib.util.spec_from_file_location(fname, modpath)
                module = importlib.util.module_from_spec(spec)
                module.vdb = vdb
                module.__file__ = modpath
                spec.loader.exec_module(module)

                if not hasattr(module, 'vdbExtension'):
                    logger.warning("Skipping python module %r, not a valid extension (module has no vdbExtension() function)", modpath)
                    continue

                logger.info("Loading extension: %r", modpath)
                module.vdbExtension(vdb, trace)
                vdb.addExtension(fname, module)

            except Exception:
                vdb.vprint('VDB Extension Error: %s' % modpath)
                vdb.vprint(traceback.format_exc())



import envi.bits as e_bits
import envi.registers as e_reg

from envi.archs.arm.const import *

'''
Strategy:
    * Use only distinct register for Register Context (unique in each bank)
    * Store a lookup table for the different banks of registers, based on the
        register data in proc_modes (see const.py)
    * Emulator does translation from register/mode to actual storage container
        using reg_table and some math (see _getRegIdx)
'''
arm_regs_tups = [
    ('r0', 32),
    ('r1', 32),
    ('r2', 32),
    ('r3', 32),
    ('r4', 32),
    ('r5', 32),
    ('r6', 32),
    ('r7', 32),
    ('r8', 32),
    ('r9', 32),
    ('r10', 32),
    ('r11', 32),
    ('r12', 32),
    ('sp', 32),  # also r13
    ('lr', 32),  # also r14
    ('pc', 32),  # also r15
    ('cpsr', 32),
    ('nil', 32),    # place holder
    ('spsr', 32),   # updated for each level
    # FIXME: need to deal with ELR_hyp
]

# force them into a tuple for faster run-time access
arm_regs_tups = tuple(arm_regs_tups)
arm_regs = [r for r,sz in arm_regs_tups]

arm_metas = [
        ("r13", REG_SP, 0, 32),
        ("r14", REG_LR, 0, 32),
        ("r15", REG_PC, 0, 32),
]

REG_PC =   15
REG_CPSR = 16
REG_NIL =  17
REG_SPSR = 18

REG_APSR_MASK = 0xffff0000

# build a translation table to allow for fast access of banked registers
modes = list(proc_modes.keys())
modes.sort()

# ARMv7/A32/T32 registers as complicated.  each processor mode accesses a particular set of 
# registers.  some regs are shared with other modes, and some are unique to a particular mode.
# to achieve this using high-performance disassembly and emulation, and the power of the Viv
# RegisterContext model, we've had to get creative.  

# the solution we've chosen is to use a big lookup table and an extra layer to access the actual
# emulated register data.  the larger list of registers (below, reg_table_data) defines a large 
# list of registers, grouped into an equal size list of register definitions per processor mode.
# the smaller list (reg_data) defines a collapsed list of registers, including only the register 
# definitions which have unique data storage.
# we use another large list (reg_table) for mapping indexes into 
# reg_table_data to indexes into reg_data

# so, thats:
# reg_table is the redirection table between MATH-COUNT-Index and Real Register Index: #index
# reg_data is that compressed "only what's needed" table of register entries:  ('name', size)
# reg_table_data is the expanded reg_data completing the all modes, etc:       ('name', size')

# other details:
# emulator get/setRegister() access the emulator data as per *reg_data*
# disassembly creates *RegOper*s using indexes into *reg_table_data*
# the ArmRegisterContext class override get/setRegister() to handle translation
# as far as the rest of the ArmRegisterContext is concern, there is only reg_data
# *RegOper* operands only care about the *reg_tables_data* since they only care about render/repr
#    and the getOperValue() defers to the ArmRegisterContext() for actually accessing the emu data

# this setup allows us to have ranges of "registers" to match the different processor modes
# where some of those registers are unique to a given mode, and others are shared (banked regs)

# all modes' CPSR's point to the one and only CPSR
# all modes` SPSR's are their own thing, and should get a copy of the current CPSR's data on 
#       mode-change.        # NOTE: THIS IS IMPORTANT FOR FULL EMULATION


# start off creating the data for the first mode (USR mode)
# and then pre-populating the reg_table and reg_table_data
reg_data = [ (reg, sz) for reg,sz in arm_regs_tups ]
reg_table = [ x for x in range(MODE_COUNT * REGS_PER_MODE) ]
reg_table_data = [ (None, 32) for x in range(MODE_COUNT * REGS_PER_MODE) ]

for idx,data in enumerate(reg_data):
    reg_table_data[idx] = data

# banked registers for different processor modes
for modenum in modes[1:]:       # skip first since we're already done
    (mname, msname, desc, offset, mode_reg_count, PSR_offset, priv_level) = proc_modes.get(modenum)
    #print("Building %s (%x): offset=0x%x   PSR_offset: 0x%x" % (msname, modenum, offset, PSR_offset))
    #import envi.interactive as ei; ei.dbg_interact(locals(), globals())
    # shared regs
    for ridx in range(mode_reg_count):
        # don't create new entries for this register, use the usr-mode reg
        reg_table[ridx+offset] = ridx

        rnm, rsz = arm_regs_tups[ridx]
        reg_table_data[ridx+offset] = ('%s_%s' % (rnm, msname), rsz) 

    # mode-regs (not including PC)
    for ridx in range(mode_reg_count, 15):
        idx = len(reg_data)
        rnm, rsz = arm_regs_tups[ridx]
        regname = rnm+"_"+msname
        reg_data.append((regname, 32))
        reg_table[ridx+offset] = idx

        reg_table_data[ridx+offset] = (regname, rsz) 

    # PC    - always the original PC
    reg_table[PSR_offset-3] = REG_PC
    reg_table_data[PSR_offset-3] = ('pc_%s' % (msname), 32)
    # CPSR  - always the original CPSR
    reg_table[PSR_offset-2] = REG_CPSR   # SPSR....??
    reg_table_data[PSR_offset-2] = ('CPSR_%s' % (msname), 32) 
    # NIL
    reg_table[PSR_offset-1] = REG_NIL
    reg_table_data[PSR_offset-1] = ('NIL_%s' % (msname), 32) 
    # PSR
    reg_table[PSR_offset] = len(reg_data)                   # -2 because we want SPSR_foo to point at the CPSR_foo.  i think?
    reg_table_data[PSR_offset] = ('SPSR_%s' % (msname), 32)
    reg_data.append(("SPSR_"+msname, 32))

# done with banked register translation table

# VFP Registers
#    VFPv2 consists of 16 doubleword registers
#    VFPv3 either 32 or 16 doubleword registers
#       VFPv3-D32 - implementation with 32 doubleword regs
#       VFPv3-D16 - implementation with 16 doubleword regs
#    VFPv4 - same as VFPv3
# Advanced SIMD and Floating point views are *not* identical
#
# we implement VFPv4-D32 since it should be backwards-compatible with all others
#  the largest accessor of this extended register bank is 128bits so we'll go with that.

REGS_VECTOR_TABLE_IDX = len(reg_table)
REGS_VECTOR_DATA_IDX = len(reg_data)
REGS_VECTOR_DELTA = REGS_VECTOR_TABLE_IDX - REGS_VECTOR_DATA_IDX

for simdreg in range(VFP_QWORD_REG_COUNT):
    simd_idx = REGS_VECTOR_TABLE_IDX + simdreg
    d = simdreg * 2
    s = d * 2
    reg_table.append(len(reg_data))
    reg_data.append(("q%d" % simdreg, 128))
    reg_table_data.append(("q%d" % simdreg, 128))
    if simdreg < 8: # VFPv4 only allows S# indexing up to S31
        arm_metas.append(("s%d" % (s),   simd_idx, 0, 32))
        arm_metas.append(("s%d" % (s+1), simd_idx, 32, 32))
        arm_metas.append(("s%d" % (s+2), simd_idx, 64, 32))
        arm_metas.append(("s%d" % (s+3), simd_idx, 96, 32))
    arm_metas.append(("d%d" % (d),   simd_idx, 0, 64))
    arm_metas.append(("d%d" % (d+1), simd_idx, 32, 64))

REG_FPSCR = len(reg_table)
reg_table.append(len(reg_data))
reg_data.append(('fpscr', 32))
reg_table_data.append(('fpscr', 32))

reg_table.append(len(reg_data))
reg_data.append(('IPSR', 32))
reg_table_data.append(('IPSR', 32))
reg_table.append(len(reg_data))
reg_data.append(('EPSR', 32))
reg_table_data.append(('EPSR', 32))
reg_table.append(len(reg_data))
reg_data.append(('PRIMASK', 32))
reg_table_data.append(('PRIMASK', 32))
reg_table.append(len(reg_data))
reg_data.append(('FAULTMASK', 32))
reg_table_data.append(('FAULTMASK', 32))
reg_table.append(len(reg_data))
reg_data.append(('BASEPRI', 32))
reg_table_data.append(('BASEPRI', 32))
reg_table.append(len(reg_data))
reg_data.append(('BASEPRI_MAX', 32))
reg_table_data.append(('BASEPRI_MAX', 32))
reg_table.append(len(reg_data))
reg_data.append(('CONTROL', 32))
reg_table_data.append(('CONTROL', 32))
reg_table.append(len(reg_data))
reg_data.append(('MSP', 32))
reg_table_data.append(('MSP', 32))
reg_table.append(len(reg_data))
reg_data.append(('PSP', 32))
reg_table_data.append(('PSP', 32))

MAX_TABLE_SIZE = len(reg_table_data)

l = locals()
e_reg.addLocalEnums(l, reg_table_data)

PSR_N = 31  # negative
PSR_Z = 30  # zero
PSR_C = 29  # carry
PSR_V = 28  # oVerflow
PSR_Q = 27
PSR_IT_BASE = 25
PSR_IT_SIZE = 10
PSR_J = 24
PSR_DNM = 20
PSR_GE = 16
PSR_E = 9
PSR_A = 8
PSR_I = 7
PSR_F = 6
PSR_T = 5
PSR_M = 0

PSR_N_bit  = 1 << PSR_N
PSR_N_mask = 0xffffffff ^ PSR_N_bit
PSR_Z_bit  = 1 << PSR_Z
PSR_Z_mask = 0xffffffff ^ PSR_Z_bit
PSR_C_bit  = 1 << PSR_C
PSR_C_mask = 0xffffffff ^ PSR_C_bit
PSR_V_bit  = 1 << PSR_V
PSR_V_mask = 0xffffffff ^ PSR_V_bit
PSR_T_bit  = 1 << PSR_T
PSR_T_mask = 0xffffffff ^ PSR_T_bit

psr_fields = [None for x in range(32)]
psr_fields[PSR_M] = "M"
psr_fields[PSR_T] = "T"
psr_fields[PSR_F] = "F"
psr_fields[PSR_I] = "I"
psr_fields[PSR_A] = "A"
psr_fields[PSR_E] = "E"
psr_fields[PSR_GE] = "GE"
psr_fields[PSR_GE+1] = "GE+1"
psr_fields[PSR_GE+2] = "GE+2"
psr_fields[PSR_GE+3] = "GE+3"
psr_fields[PSR_DNM] = "DNM"
psr_fields[PSR_DNM+1] = "DNM+1"
psr_fields[PSR_DNM+2] = "DNM+2"
psr_fields[PSR_DNM+3] = "DNM+3"
psr_fields[PSR_J] = "J"
psr_fields[PSR_IT_BASE] = "IT_BASE"
psr_fields[PSR_IT_BASE+1] = "IT_BASE+1"
psr_fields[PSR_IT_BASE+2] = "IT_BASE+2"
psr_fields[PSR_IT_SIZE] = "IT_SIZE"  # IT is split into two sections
psr_fields[PSR_IT_SIZE+1] = "IT_SIZE+1"  # IT is split into two sections
psr_fields[PSR_IT_SIZE+2] = "IT_SIZE+2"  # IT is split into two sections
psr_fields[PSR_IT_SIZE+3] = "IT_SIZE+3"  # IT is split into two sections
psr_fields[PSR_IT_SIZE+4] = "IT_SIZE+4"  # IT is split into two sections
psr_fields[PSR_Q] = "Q"
psr_fields[PSR_V] = "V"
psr_fields[PSR_C] = "C"
psr_fields[PSR_Z] = "Z"
psr_fields[PSR_N] = "N"

arm_status_metas = [
        ("N", REG_FLAGS, PSR_N, 1, "Negative/LessThan flag"),
        ("Z", REG_FLAGS, PSR_Z, 1, "Zero flag"),
        ("C", REG_FLAGS, PSR_C, 1, "Carry/Borrow/Extend flag"),
        ("V", REG_FLAGS, PSR_V, 1, "oVerflow flag"),
        ("Q", REG_FLAGS, PSR_Q, 1, "Sticky Overflow flag"),
        ("J", REG_FLAGS, PSR_J, 1, "Jazelle Mode bit"),
        ("GE",REG_FLAGS, PSR_GE, 4, "Greater/Equal flag"),
        ("DNM",REG_FLAGS, PSR_DNM, 4, "DO NOT MODIFY bits"),
        ("IT_BASE0",REG_FLAGS, PSR_IT_BASE, 1, "IfThen 0 bit"),
        ("IT_BASE1",REG_FLAGS, PSR_IT_BASE+1, 1, "IfThen 1 bit"),
        ("IT_BASE2",REG_FLAGS, PSR_IT_BASE+2, 1, "IfThen 2 bit"),
        ("IT_SIZE0",REG_FLAGS, PSR_IT_SIZE, 1, "IfThen 0 bit"),
        ("IT_SIZE1",REG_FLAGS, PSR_IT_SIZE+1, 1, "IfThen 1 bit"),
        ("IT_SIZE2",REG_FLAGS, PSR_IT_SIZE+2, 1, "IfThen 2 bit"),
        ("IT_SIZE3",REG_FLAGS, PSR_IT_SIZE+3, 1, "IfThen 3 bit"),
        ("IT_SIZE4",REG_FLAGS, PSR_IT_SIZE+4, 1, "IfThen 4 bit"),
        ('IT_SIZE',REG_FLAGS, PSR_IT_SIZE, 5, "IfThen Block Size"),
        ('IT_BASE',REG_FLAGS, PSR_IT_BASE, 3, "IfThen Base Condition"),
        ("E", REG_FLAGS, PSR_E, 1, "Data Endian bit"),
        ("A", REG_FLAGS, PSR_A, 1, "Imprecise Abort Disable bit"),
        ("I", REG_FLAGS, PSR_I, 1, "IRQ disable bit"),
        ("F", REG_FLAGS, PSR_F, 1, "FIQ disable bit"),
        ("T", REG_FLAGS, PSR_T, 1, "Thumb Mode bit"),
        ("M", REG_FLAGS, PSR_M, 5, "Processor Mode"),
        ]

e_reg.addLocalStatusMetas(l, arm_metas, arm_status_metas, "CPSR")
e_reg.addLocalMetas(l, arm_metas)

def getRegDataIdx(idx):
    ridx = reg_table[idx]  # magic pointers allowing overlapping banks of registers
    return ridx

class ArmRegisterContext(e_reg.RegisterContext):
    def __init__(self):
        e_reg.RegisterContext.__init__(self)
        self.loadRegDef(reg_table_data)
        self.loadRegMetas(arm_metas, statmetas=arm_status_metas)
        self.setRegisterIndexes(REG_PC, REG_SP)

    def getProcMode(self):
        '''
        get current ARM processor mode.  see proc_modes (const.py)
        '''
        return self._rctx_vals[REG_CPSR] & 0x1f

    def setProcMode(self, mode):
        '''
        set the current processor mode.  stored in CPSR
        '''
        # write current psr to the saved psr register for target mode
        # but not for USR or SYS modes, which don't have their own SPSR
        if mode not in (PM_usr, PM_sys):
            curSPSRidx = proc_modes[mode]
            self._rctx_vals[curSPSRidx] = self.getCPSR()

        # set current processor mode
        cpsr = self._rctx_vals[REG_CPSR] & 0xffffffe0
        self._rctx_vals[REG_CPSR] = cpsr | mode

    def getRegister(self, index, mode=None):
        """
        Return the current value of the specified register index.
        """
        if mode is None:
            mode = self.getProcMode() & 0xf
        else:
            mode &= 0xf

        idx = (index & 0xffff)
        ridx = _getRegIdx(idx, mode)

        if (index & 0xffff) == index:   # not a metaregister
            return self._rctx_vals[ridx]

        offset = (index >> 24) & 0xff
        width  = (index >> 16) & 0xff

        mask = (2**width)-1
        return (self._rctx_vals[ridx] >> offset) & mask

    def setRegister(self, index, value, mode=None):
        """
        Set a register value by index.
        """
        if mode is None:
            mode = self.getProcMode() & 0xf
        else:
            mode &= 0xf

        self._rctx_dirty = True

        # the raw index (in case index is a metaregister)
        idx = (index & 0xffff)

        # have to translate every register index to find the right home
        ridx = _getRegIdx(idx, mode)

        if (index & 0xffff) == index:   # not a metaregister
            self._rctx_vals[ridx] = (value & self._rctx_masks[ridx])      # FIXME: hack.  should look up index in proc_modes dict?
            return

        # If we get here, it's a meta register index.
        # NOTE: offset/width are in bits...
        offset = (index >> 24) & 0xff
        width  = (index >> 16) & 0xff

        mask = e_bits.b_masks[width]
        mask = mask << offset

        # NOTE: basewidth is in *bits*
        basewidth = self._rctx_widths[ridx]
        basemask  = (2**basewidth)-1

        # cut a whole in basemask at the size/offset of mask
        finalmask = basemask ^ mask

        curval = self._rctx_vals[ridx]

        self._rctx_vals[ridx] = (curval & finalmask) | (value << offset)


def _getRegIdx(idx, mode):
    if idx >= REGS_VECTOR_TABLE_IDX:
        return reg_table[idx]

    mode &= 0xf
    ridx = idx + (mode*REGS_PER_MODE)  # account for different banks of registers
    ridx = reg_table[ridx]  # magic pointers allowing overlapping banks of registers
    return ridx


def reg_mode_base(mode):
    '''
    Get the base register index for a given mode (full-sized table)
    '''
    mdata = proc_modes.get(mode)
    if mdata is None:
        raise Exception("Invalid Mode access (reg_mode_base): %d" % mode)

    return mdata[3]



rctx = ArmRegisterContext()



import vtrace
import vstruct
import vivisect.impapi as viv_impapi

class StealthBreak(vtrace.Breakpoint):
    '''
    Base class that can be extended by other classes to bypass anti-debugging
    checks.
    '''
    def __init__(self, sym=None):
        vtrace.Breakpoint.__init__(self, None, expression=sym)
        self.fastbreak = True

        self.bpid = None
        self.ispatched = False
        self.name = None

        self.impapi = None
        self.cc = None
        self.argc = None

    def isPatched(self):
        return self.ispatched

    def getName(self):
        return self.name

    def enablePatch(self, trace):
        if not self.isPatched():
            self.bpid = trace.addBreakpoint(self)
            self.ispatched = True

    def disablePatch(self, trace):
        if self.isPatched():
            trace.removeBreakpoint(self.bpid)
            self.ispatched = False
            self.bpid = None

    def notify(self, event, trace):
        if self.impapi is not None:
            # cached
            return

        self.impapi = viv_impapi.getImportApi(trace.getMeta('Platform'), trace.getMeta('Architecture'))
        cc = self.impapi.getImpApiCallConv(self.vte)
        emu = vtrace.getEmu(trace)
        self.cc = emu.getCallingConvention(cc)
        self.argc = len(self.impapi.getImpApiArgs(self.vte))

class StealthPeb(StealthBreak):
    '''
    Disables the "BeingDebugged" and "NtGlobalFlag" flags in the PEB.
    Also modifies heap flags that indicate debugging.
    '''
    def __init__(self):
        self.patched = False
        self.savedForceFlags = None
        self.savedNtGlobalFlags = None
        self.name = 'Peb'

    def isPatched(self):
        return self.patched

    def writeBeingDebugged(self, trace, val):
        peb = trace.parseExpression('peb')
        ps = vstruct.getStructure('win32.PEB')
        off = ps.vsGetOffset('BeingDebugged')
        trace.writeMemoryFormat(peb+off, '<B', val)
        #TODO Save off the old peb so we can fix it when we want to break
        #trace.setMeta("Win32StealthPeb", val)

    def writeProcessHeapFlags(self, trace, val):
        pebaddr = trace.parseExpression('peb')
        peb = trace.getStruct('win32.PEB', pebaddr)
        structHeap = trace.getStruct('win32.HEAP', peb.ProcessHeap)
        self.savedForceFlags = structHeap.ForceFlags
        structHeap.ForceFlags = val
        trace.writeMemory(peb.ProcessHeap, structHeap.vsEmit())

    def writeNtGlobalFlag(self, trace, val):
        pebaddr = trace.parseExpression('peb')
        peb = trace.getStruct('win32.PEB', pebaddr)
        self.savedNtGlobalFlags = peb.NtGlobalFlag
        peb.NtGlobalFlag = val
        trace.writeMemory(pebaddr, peb.vsEmit())

    def enablePatch(self, trace):
        trace.requireNotRunning()
        self.patched = True
        self.writeBeingDebugged(trace, 0)
        self.writeProcessHeapFlags(trace, 0)
        self.writeNtGlobalFlag(trace, 0)

    def disablePatch(self, trace):
        trace.requireNotRunning()
        self.patched = False
        self.writeBeingDebugged(trace, 1)
        self.writeProcessHeapFlags(trace, self.savedForceFlags)
        self.writeNtGlobalFlag(trace, self.savedNtGlobalFlags)

class StealthCheckRemoteDebuggerPresent(StealthBreak):
    '''
    Forces the "CheckRemoteDebuggerPresent" API to indicate that the process
    is not being debugged.
    '''
    def __init__(self, sym):
        StealthBreak.__init__(self, sym)
        self.name = 'CheckRemoteDebuggerPresent'

    def notify(self, event, trace):
        StealthBreak.notify(self, event, trace)
        handle, outbool = self.cc.getCallArgs(trace, self.argc)
        trace.writeMemoryFormat(outbool, '<P', 0)
        self.cc.execCallReturn(trace, 1, self.argc)

        trace.runAgain()

class StealthGetTickCount(StealthBreak):
    '''
    Returns a static tickcount in case the application checks time deltas
    between instructions.
    '''
    def __init__(self, sym):
        StealthBreak.__init__(self, sym)
        self.name = 'GetTickCount'
        self.tickReturn = 0xdeadbeef

    def notify(self, event, trace):
        StealthBreak.notify(self, event, trace)
        self.cc.execCallReturn(trace, self.tickReturn, self.argc)
        self.tickReturn += 1

        trace.runAgain()

class StealthOutputDebugString(StealthBreak):
    '''
    Forces the OutputDebugString API to return 1.
    '''
    def __init__(self, sym):
        StealthBreak.__init__(self, sym)
        self.name = 'OutputDebugString'

    def notify(self, event, trace):
        StealthBreak.notify(self, event, trace)
        self.cc.execCallReturn(trace, 1, self.argc)

        trace.runAgain()

class StealthZwClose(StealthBreak):
    '''
    When called with a invalid handle (-1) the malware will simply return so
    an exception is not thrown when a debugger is attached.
    '''
    def __init__(self, sym):
        StealthBreak.__init__(self, sym)
        self.name = 'ZwClose'

    def notify(self, event, trace):
        StealthBreak.notify(self, event, trace)
        handle = self.cc.getCallArgs(trace, self.argc)
        if handle == -1:
            self.cc.execCallReturn(trace, 0, self.argc)

        trace.runAgain()

class StealthZwSetInformationThread(StealthBreak):
    '''
    When ZwSetInformationThread is called with ThreadHideFromDebugger, just
    return in case threads try to detach from the debugger.
    '''
    def __init__(self, sym):
        StealthBreak.__init__(self, sym)
        self.name = 'ZwSetInformationThread'

    def notify(self, event, trace):
        StealthBreak.notify(self, event, trace)
        handle, threadInfoClass, threadInfo, threadInfoLen = self.cc.getCallArgs(trace, self.argc)
        # ThreadHideFromDebugger = 0x11
        if threadInfoClass == 0x11:
            # If ThreadHideFromDebugger is passed, fake it
            trace.writeMemoryFormat(threadInfo, '<P', 0)
            self.cc.execCallReturn(trace, 0, self.argc)

        trace.runAgain()

class StealthZwQueryInformationProcess(StealthBreak):
    '''
    Forces ZwQueryInformationProcess to return success when passed
    "ProcessDebugPort"
    '''
    def __init__(self, sym):
        StealthBreak.__init__(self, sym)
        self.name = 'ZwQueryInformationProcess'

    def notify(self, event, trace):
        StealthBreak.notify(self, event, trace)
        handle, procInfoClass, procInfo, procInfoLen, retLen = self.cc.getCallArgs(trace, self.argc)
        # ProcessDebugPort = 7
        if procInfoClass == 7:
            # If ProcessDebugPort is passed, fake it out by setting
            # ProcessInformation to zero
            trace.writeMemoryFormat(procInfo, '<P', 0)
            trace.writeMemoryFormat(retLen, '<P', procInfoLen)
            self.cc.execCallReturn(trace, 0, self.argc)

        trace.runAgain()

def stealthInit(trace):
    objs = []
    objs.append(StealthPeb())
    objs.append(StealthCheckRemoteDebuggerPresent('kernel32.CheckRemoteDebuggerPresent'))
    objs.append(StealthGetTickCount('kernel32.GetTickCount'))
    objs.append(StealthZwClose('ntdll.ZwClose'))
    objs.append(StealthZwSetInformationThread('ntdll.ZwSetInformationThread'))
    objs.append(StealthOutputDebugString('kernel32.OutputDebugStringA'))
    objs.append(StealthZwQueryInformationProcess('ntdll.ZwQueryInformationProcess'))

    trace.setMeta('Win32Stealth', objs)

def getStatus(trace):
    statList = []
    if not trace.getMeta('Win32Stealth'):
       stealthInit(trace)

    stealthObjs = trace.getMeta('Win32Stealth')
    for inst in stealthObjs:
        statList.append((inst.name, inst.isPatched()))

    return statList

def enableAllStealth(trace):
    if not trace.getMeta('Win32Stealth'):
       stealthInit(trace)

    stealthObjs = trace.getMeta('Win32Stealth')
    for inst in stealthObjs:
        inst.enablePatch(trace)

def disableAllStealth(trace):
    if not trace.getMeta('Win32Stealth'):
       stealthInit(trace)

    stealthObjs = trace.getMeta('Win32Stealth')
    for inst in stealthObjs:
        inst.disablePatch(trace)

def stealthify(trace, name):
    if not trace.getMeta('Win32Stealth'):
       stealthInit(trace)

    stealthObjs = trace.getMeta('Win32Stealth')
    for inst in stealthObjs:
        if inst.getName().lower() == name:
            inst.enablePatch(trace)

def unstealthify(trace, name):
    if not trace.getMeta('Win32Stealth'):
       stealthInit(trace)

    stealthObjs = trace.getMeta('Win32Stealth')
    for inst in stealthObjs:
        if inst.getName().lower() == name:
            inst.disablePatch(trace)
